%{
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef KERNEL_MODULE
#include <configfile.h>
#endif

#include <stdlib.h>	/* strtol() */
#include <string.h>	/* strncpy() */
#include <stdio.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#include "adconfig_yacc.h"
static void usage(void);
static void usage_add(void);
static void usage_del(void);
static void usage_config(void);
static void list(void);

#include <linux/adwin.h>

int psc = 0;		/* Previous Start Condition */

extern char *progname;
extern void version();

%}
/* don't use multiple input buffers */
%option noyywrap


dec		[0-9]+
oct		"0"([0-7]{1,3})
hex		"0x"[0-9a-f]+

ipseg		[0-9]{1,3}
ipv4		{ipseg}"."{ipseg}"."{ipseg}"."{ipseg}

macseg		[a-f0-9]{2}
mac		{macseg}":"{macseg}":"{macseg}":"{macseg}":"{macseg}":"{macseg}

/* exclusive start conditions */
%x help add del config devtype _link net usb
%x ip mac dec hex oct string bool
%x deviceno uid gid
%%

<bool>0|false|no			{ yylval.val = 0;  BEGIN(psc); return VAL; }
<bool>1|true|yes			{ yylval.val = 1;  BEGIN(psc); return VAL; }
<bool><<EOF>>				{ fprintf(stderr, "error: boolean value expected\n"); return FLEXERR; }
<bool>.*				{ fprintf(stderr, "error: expected boolean value, got \"%s\"\n", yytext); return FLEXERR; }

<oct>{oct}				{ yylval.val = strtol(yytext, NULL, 8); BEGIN(psc); return VAL; }
<oct><<EOF>>				{ fprintf(stderr, "error: octal value expected\n"); return FLEXERR; }
<oct>.*					{ fprintf(stderr, "error: expected octal value, got \"%s\"\n", yytext); return FLEXERR; }

<dec>{dec}				{ yylval.val = strtol(yytext, NULL, 10); BEGIN(psc); return VAL; }
<dec><<EOF>>				{ fprintf(stderr, "error: decimal value expected\n"); return FLEXERR; }
<dec>.*					{ fprintf(stderr, "error: expected decimal value, got \"%s\"\n", yytext); return FLEXERR; }

<hex>{hex}				{ yylval.val = strtol(yytext, NULL, 16); BEGIN(psc); return VAL; }
<hex><<EOF>>				{ fprintf(stderr, "error: hexadecimal value expected\n"); return FLEXERR; }
<hex>.*					{ fprintf(stderr, "error: expected hexadecimal value, got \"%s\"\n", yytext); return FLEXERR; }

<deviceno>{dec}|{hex}			{ yylval.val = strtol(yytext, NULL, 0); 
                                          if ((yylval.val < 1) || (yylval.val > 0xfff)) {
                                                  fprintf(stderr, "error: device number must be > 0 and < 0x1000\n"); 
                                                  return FLEXERR;
                                          }

                                          BEGIN(psc); return VAL; 
                                        }
<deviceno><<EOF>>			{ fprintf(stderr, "error: expected device number\n"); return FLEXERR; }
<deviceno>.*				{ fprintf(stderr, "error: expected (hexa-)decimal value, got \"%s\"\n", yytext); return FLEXERR; }

<uid,gid>{dec}				{ yylval.val = strtol(yytext, NULL, 10); BEGIN(psc); return VAL; }
<uid><<EOF>>				{ fprintf(stderr, "error: expected user id\n"); return FLEXERR; }
<gid><<EOF>>				{ fprintf(stderr, "error: expected group id\n"); return FLEXERR; }
<uid,gid>.*				{ strncpy(yylval.str, yytext, 40); BEGIN(psc); return STR; }

<string><<EOF>>				{ fprintf(stderr, "error: string expected\n"); return FLEXERR; }
<string>.*				{ strncpy(yylval.str, yytext, 40); BEGIN(psc); return STR; }

<ip>{ipv4}				{ struct in_addr addr;
					  if (! inet_aton(yytext, &addr)) {
						fprintf(stderr, "error: invalid IP \"%s\"\n", yytext);
						return FLEXERR;
					  }
					  strncpy(yylval.str, yytext, 16);
					  BEGIN(psc);
					  return IPSTR;
					}
<ip><<EOF>>				{ fprintf(stderr, "error: IP expected\n"); return FLEXERR; }
<ip>.*					{ fprintf(stderr, "error: expected valid IP, got \"%s\"\n", yytext); return FLEXERR; }

<mac>{mac}				{ strncpy(yylval.str, yytext, 18); BEGIN(psc); return MACSTR; }
<mac><<EOF>>				{ fprintf(stderr, "error: hardware address (MAC) expected\n"); return FLEXERR; }
<mac>.*					{ fprintf(stderr, "error: expected valid hardware address (MAC), got \"%s\"\n", yytext); return FLEXERR; }



add					{ psc = add; BEGIN(deviceno); return ADD; }
del					{ psc = del; BEGIN(deviceno); return DEL; }
config					{ psc = config; BEGIN(mac); return CONFIG; }
--help					{ BEGIN(help); return HELP; }
--version				{ version(); }
<INITIAL><<EOF>>			{ list(); return FLEXERR; }
.*					{ fprintf(stderr, "error: invalid command \"%s\"\n", yytext);
					  return FLEXERR;
					}

<help>add				{ usage_add(); return HELP; }
<help>del				{ usage_del(); return HELP; }
<help>config				{ usage_config(); return HELP; }
<help><<EOF>>				{ usage(); return HELP; }
<help>.*				{ usage(); return FLEXERR; }


<del>.*					{ fprintf(stderr, "error: invalid argument \"%s\".\n", yytext);
					  fprintf(stderr, "Use \"%s --help del\" to get further information.\n", progname);
					  return FLEXERR;
					}


<add,_link,net,usb>uid			{ psc = YY_START; BEGIN(uid); return UID; }
<add,_link,net,usb>gid			{ psc = YY_START; BEGIN(gid); return GID; }
<add,_link,net,usb>mode			{ psc = YY_START; BEGIN(oct); return MODE; }
<add,_link,net,usb>desc			{ psc = YY_START; BEGIN(string); return DESC; }


<add>type				{ BEGIN(devtype); return TYPE; }
<add>.*					{ fprintf(stderr, "error: invalid  argument \"%s\".\n", yytext);
					  fprintf(stderr, "Use \"%s --help add\" to get further information.\n", progname);
					  return FLEXERR;
					}

<devtype>link 				{ yylval.val = DC_LINK; BEGIN(_link); return DEVTYPE; }
<devtype>net 				{ yylval.val = DC_ETHERNET;  BEGIN(net); return DEVTYPE; }
<devtype>usb 				{ yylval.val = DC_USB;  BEGIN(usb); return DEVTYPE; }
<devtype><<EOF>>			{ fprintf(stderr, "error: device type expected\n");
					  return FLEXERR;
					}
<devtype>.*				{ fprintf(stderr, "error: Invalid device type \"%s\".\n", yytext);
					  return FLEXERR;
					}

<_link>base				{ psc = YY_START; BEGIN(hex); return BASE; }
<_link>.*				{ fprintf(stderr, "error: invalid argument \"%s\" in context `add type LINK`\n", yytext);
					  return FLEXERR;
					}

<usb>serial				{ psc = YY_START; BEGIN(dec); return SERIAL; }
<usb>.*					{ fprintf(stderr, "error: invalid argument \"%s\" in context `add type USB`\n", yytext);
					  return FLEXERR;
					}

<net>port				{ psc = YY_START; BEGIN(dec); return PORT; }
<net>hostlink				{ psc = YY_START; BEGIN(hex); return HOSTLINK; }
<net>timeout				{ psc = YY_START; BEGIN(dec); return TIMEOUT; }
<net>count				{ psc = YY_START; BEGIN(dec); return COUNT; }
<net>ip					{ psc = YY_START; BEGIN(ip); return IP; }
<net>pw					{ psc = YY_START; BEGIN(string); return PW; }
<net>.*					{ fprintf(stderr, "error: invalid argument \"%s\" in context `add type NET`\n", yytext);
					  return FLEXERR;
					}

<config>port				{ psc = YY_START; BEGIN(dec); return PORT; }
<config>pw				{ psc = YY_START; BEGIN(string); return PW; }
<config>ip				{ psc = YY_START; BEGIN(ip); return IP; }
<config>mask				{ psc = YY_START; BEGIN(ip); return MASK; }
<config>gateway				{ psc = YY_START; BEGIN(ip); return GATEWAY; }
<config>dhcp				{ psc = YY_START; BEGIN(bool); return DHCP; }
<config>.*				{ fprintf(stderr, "error: invalid argument \"%s\".\n", yytext);
					  fprintf(stderr, "Use \"%s --help config\" to get further information.\n", progname);
					  return FLEXERR;
					}

%%

#define HELPLINE(option, text)		fprintf(stderr, "\t%-20s%s\n", option, text);

static void usage(void) {
	fprintf(stderr, "%s - ADwin system setup for LINUX\n", progname);
	fprintf(stderr, "(c) %s Jaeger Messtechnik GmbH, Lorsch\n", COPYRIGHT_YEARS);
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "    %s <command> [OPTIONS]... \n", progname);
	fprintf(stderr, "    %s without arguments displays all configured devices.\n", progname);
	fprintf(stderr, "\n");
	fprintf(stderr, "Remark:\n");
	fprintf(stderr, "    Device numbers are HEX values!\n");
	fprintf(stderr, "    From 'add 12 ...' you will get the device '0xC'. Use 'add 0x12' instead.\n");
	fprintf(stderr, "\n");

	HELPLINE("add", "Add a new deviceno");
	HELPLINE("del", "Remove a given deviceno");
	HELPLINE("config", "Configure the system owning a given hwaddr");
	HELPLINE("", "");
	HELPLINE("--help", "This text. Try \"--help XXX\" for more");
	HELPLINE("", "detailed information about command XXX");
	HELPLINE("--version", "Print version information");
	HELPLINE("", "");

	fprintf(stderr, "Report bugs and suggestions to <info@adwin.de>\n");
}

static void usage_add(void) {
	fprintf(stderr, "Add a new ADwin deviceno.\n");
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "    %s add <deviceno> [UID <uid|username>] [GID <gid|groupname>]\n", progname);
	fprintf(stderr,	"    [MODE <mode>] [DESC <\"\">] <TYPE [link <BASE <#>>] [usb <SERIAL <#>>]\n");
	fprintf(stderr,	"     [net <IP <addr>> [PW <\"\">] [PORT <#>] [TIMEOUT <#>] [COUNT <#>]\n");
	fprintf(stderr,	"      [HOSTLINK <#>]");
	fprintf(stderr, "]>\n\n");

	HELPLINE("UID <uid|username>", "Owner of the device (default: root)");
	HELPLINE("GID <gid|groupname>", "Group of new device (default: root)");
	HELPLINE("MODE <mode>", "Set the file access mode (default: 0666)");
	HELPLINE("DESC <\"\">", "Description string (32 char. max, default: \"\")");

	HELPLINE("", "");

	HELPLINE("TYPE", "One of the following types must be chosen:");
	HELPLINE("  [link]", "  Creates a new link device");
	HELPLINE("    BASE <#>", "    Required, link base address (e.g. 0x150)");
	HELPLINE("  [usb]", "  Creates a new USB device");
	HELPLINE("    SERIAL <#>", "    Required, USB serial number");
	HELPLINE("  [net]", "  Creates a new Ethernet device");
	HELPLINE("    IP <addr>", "    Required, network address");
	HELPLINE("    PW <\"\">", "    System password (default: \"\")");
	HELPLINE("    PORT <#>", "    Network port number (default: 6543)");
	HELPLINE("    TIMEOUT <#>", "    Protocol timeout (default: 1000ms)");
	HELPLINE("    COUNT <#>", "    USP package sequence count (default: 5)");
	HELPLINE("    HOSTLINK <#>", "    Only necessary if a software server is used");
	HELPLINE("", "    (default: 0x0)");
	HELPLINE("", "");
}

static void usage_del(void) {
	fprintf(stderr, "Remove a given ADwin deviceno.\n");
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "    %s del <deviceno>\n", progname);
	fprintf(stderr, "\n");
}

static void usage_config(void) {
	fprintf(stderr, "Configure the ADwin Ethernet Device owning a given hwaddr.\n");
	fprintf(stderr, "Since the broadcast address 255.255.255.255 is used to configure,\n");
	fprintf(stderr, "Please make sure the ADwin System is attached to a network reachable\n");
	fprintf(stderr, "from the first network interface (eth0). Otherwise config will NOT\n");
	fprintf(stderr, "work!\n");
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "    %s config <hwaddr> [IP <addr>] [MASK <addr>] [GATEWAY <addr>]\n"
			"    [PORT <#>] [DHCP <yes|no>] [PW <\"\">]\n", progname);
	fprintf(stderr, "\n");

	HELPLINE("IP <addr>", "The new IP the system shall use");
	HELPLINE("",   "(required if not \"dhcp yes\")");
	HELPLINE("MASK <addr>", "Subnet mask (required if not \"dhcp yes\"");

	HELPLINE("GATEWAY <addr>", "Default gateway (default: 0.0.0.0, no gateway)");
	HELPLINE("PORT <#>", "Network port number (default: 6543)");
	HELPLINE("PW <\"\">", "New password to set");
	HELPLINE("", "(default: \"\", existing PW will be removed!)");
	HELPLINE("DHCP <no|yes>", "Use a DHCP server to obtain an addr (default: no)");
	HELPLINE("", "Remark: DHCP usage is not recommended.");
	HELPLINE("", "");
}

static void list(void) {
	char line[255];

	FILE *proc = fopen(CONFIG_FILE, "r");

	if (! proc) {
		perror(CONFIG_FILE);
		return;
	}

	while(fgets(line, 255, proc) > 0)
		fprintf(stdout, line);

	fprintf(stdout, "\n");
	fclose(proc);
}
