/***************************************************************************
                          ADEthernet.c  -  description
                             -------------------
    begin               : Wed Apr 4 2001
    copyright           : (C) 2001 by Jaeger Computergesteuerte Messtechnik GmbH
    email               : info@adwin.de
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include <adwin-macros.h>

#if !defined (WIN32) && !defined (__BORLANDC__) /* Linux */
#  include <unistd.h>
#  include <sys/socket.h>
#  include <sys/un.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <sys/time.h>
#  include <pthread.h>

#  define CLOSE(x) close (x)
#  define SLEEP(x) usleep (x * 1000)
#  define SOCKET_TYPE int
#  define RESET_TIMEOUT(t) {(t).tv_sec = D->timeout / 1000; \
		                (t).tv_usec = (D->timeout % 1000) * 1000; }
#else /* Windows */
#  include <winsock2.h>

#  define CLOSE(x) closesocket(x)
#  define SLEEP(x) Sleep (x)
#  define SOCKET_TYPE SOCKET
#  define RESET_TIMEOUT(t)
#endif	// !defined (WIN32) && !defined (__BORLANDC__)

// adwin-includes
#include <adwin-macros.h>
#include "libadwin/bits.h"
#include "libadwin.h"
#include "retry_count.h"
#include "UDPADwin.h"
#include "getDataSH.h"
#include "ADEthernet.h"
#include "FehlerText.h"
#include "libadwin-macros.h"
#include "data_convert.h"

// Parameter, die zu einem ADwin-System gehren (ADEthernet.c kennt nur 1 Device pro Thread gleichzeitig)
typedef struct
{
	SOCKET_TYPE sockfd;
	struct sockaddr_in client, server;
	unsigned int timeout;
	char passwd[10];
	int hostlink;
	int32_t fehlercode;
	uint32_t udp_paket_count;
	int32_t ip_version;
} ADwinDevice;

short setarray (unsigned char *arr, int32_t msize, int32_t typsize,SOCKET_TYPE msocket, struct sockaddr *server, int max_paket_count, struct timeval *tv, int swap_bytes);
short getarray (char *arr, int32_t msize, int32_t typsize,SOCKET_TYPE msocket, struct sockaddr *m_serv_addr, int max_paket_count, struct timeval *tv);

#if !defined (WIN32) && !defined (__BORLANDC__)

	static pthread_once_t DeviceOnceControl = PTHREAD_ONCE_INIT;
	static pthread_key_t DeviceKey;

	void FreeDeviceKey (void *arg)
	{
		CLOSE (((ADwinDevice*)arg)->sockfd);
		if (arg) free (arg);
	}

	void InitDeviceKey ()
	{
		pthread_key_create (&DeviceKey, FreeDeviceKey);
	}

	static ADwinDevice* GetCurrentDevice ()
	{
		ADwinDevice *CurrentDevice;

		pthread_once(&DeviceOnceControl, InitDeviceKey);
		CurrentDevice = pthread_getspecific (DeviceKey);
		if (!CurrentDevice) {
			CurrentDevice = malloc(sizeof (ADwinDevice));
			pthread_setspecific(DeviceKey, CurrentDevice);
		}
		return CurrentDevice;
	}

	#define D (GetCurrentDevice ())
#else
	ADwinDevice CurrentDevice;

	#define D (&CurrentDevice)
#endif

static int32_t intern_Get_Version_TcpIp();
static short Get_Data_fastshort(int32_t, void*, short, int32_t, int32_t);

static void getDataClassic(int32_t data_no, void *data, short type,
			   int32_t start_index, int32_t count);
static void getDataSH(int32_t timeout, void *data, short type, 
		      short data_no, int32_t start_index, int32_t count);
// Globale Variablen

short english=1;

#define DLLVERSION 3000445
#define PLATFORM 0x80

int openConnect(const char *IP, const char *password,
		int link, int Port, unsigned int UDPTimeout)
{
	int lauf, serversize, tmp;
#if defined (WIN32) || defined (__BORLANDC__)
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 2 );

	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 )
	{
		return 1;
	}

	if ( LOBYTE( wsaData.wVersion ) != 2 || HIBYTE( wsaData.wVersion ) != 2 )
	{
	    WSACleanup( );
		return 1;
	}
#endif	// defined (WIN32) || defined (__BORLANDC__)
	PRINTDEBUG(DBG_GENERIC_DRV, 
		   ("openConnect Params: IP = %s, password = %s, link = %d\n"
		    "Port = %d, Timeout = %d\n",
		    IP, password, link, Port, UDPTimeout));

	// timeouts
	D->timeout  =  UDPTimeout;

	// Passwort uebernehmen
	strncpy(D->passwd, password, 10);

	// Link
	D->hostlink = link;

	// Socket kreieren
	D->sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (D->sockfd == -1) return(1);
	PRINTDEBUG(DBG_GENERIC_DRV, ("socketfd is %d\n", D->sockfd));

	// Paket-Zhler initialisieren
	D->udp_paket_count = 1;

	// Angabe zum Client
	D->client.sin_family = AF_INET;
	D->client.sin_port = 0;//htons(Port);
	D->client.sin_addr.s_addr = INADDR_ANY;

	// Angaben zum Server
	D->server.sin_family = AF_INET;
#if !defined (WIN32) && !defined (__BORLANDC__)
	bzero(&(D->server.sin_zero), 8);
	inet_aton(IP, &D->server.sin_addr);
	D->server.sin_port = htons((in_port_t) Port);
#else
	ZeroMemory (&(D->server.sin_zero), 8);
	D->server.sin_addr.S_un.S_addr = inet_addr (IP);
	D->server.sin_port = htons((short) Port);
	if (D->server.sin_addr.S_un.S_addr == INADDR_NONE) return 1;
#endif
	serversize = sizeof(D->server);

	PRINTDEBUG(DBG_GENERIC_DRV, ("setting socket options\n"));
	lauf = 100000;
	tmp = setsockopt(D->sockfd, SOL_SOCKET , SO_SNDBUF,
			 (const unsigned char *)&lauf,sizeof(int));
	if (tmp == -1) 
		perror("openConnect, setcockopt, SND");

	lauf = 100000;
	tmp = setsockopt(D->sockfd, SOL_SOCKET, SO_RCVBUF,
			 (const unsigned char *)&lauf,sizeof(int));
	if (tmp == -1) 
		perror("openConnect, setcockopt, RCV");

	D->ip_version = 0; // D->ip_version is used by
			   // intern_Get_Version_TcpIP an valgrind
			   // will issue a warning for the following
			   // call if ip_version is not initialized
	D->ip_version = intern_Get_Version_TcpIp();

	if (! D->ip_version) {
		return 1;
	}
	D->fehlercode = 0;
	PRINTDEBUG(DBG_GENERIC_DRV, ("end openConnect\n"));
	return(0);
}

int reConnect (const char *IP, const char *password, int link, 
	       int Port, unsigned int UDPTimeout)
{

	PRINTDEBUG(DBG_GENERIC_DRV, 
		   ("reConnect mit IP = %s, password = %s, link = %d\n", 
		   IP, password, link));

	// timeouts
	D->timeout = UDPTimeout;

	// Passwort uebernehmen
	strncpy(D->passwd, password, 10);

	// Link
	D->hostlink = link;

	// Paket-Zhler initialisieren
	D->udp_paket_count = 1;

#if !defined (WIN32) && !defined (__BORLANDC__)
	bzero(&(D->server.sin_zero), 8);
	inet_aton(IP, &D->server.sin_addr);
	D->server.sin_port = htons((in_port_t)Port);
#else
	ZeroMemory (&(D->server.sin_zero), 8);
	D->server.sin_addr.S_un.S_addr = inet_addr (IP);
	D->server.sin_port = htons((short)Port);
	if (D->server.sin_addr.S_un.S_addr == INADDR_NONE) return 1;
#endif
	// TODO: is this necessary?
	D->ip_version = intern_Get_Version_TcpIp();
	if (! D->ip_version) {
		return 1;
	}
	D->fehlercode = 0;
	PRINTDEBUG(DBG_GENERIC_DRV, ("end reConnect\n"));
	return(0);
}

void closeConnect()
{
	// Socket schliessen
	CLOSE (D->sockfd);

#if defined (WIN32) || defined (__BORLANDC__)
	WSACleanup ();
#endif
}

//TODO: hier kann man doch FD_ZERO,etc weglassen, oder? ansonsten OK
int sendtoServer(int msocket, struct sockaddr_in *server, 
		 UDPH1 *h1,struct timeval *tv)
{
	int res = k_UDPtimeout;
	int count, tmp;
	fd_set rset;
	
	PRINTDEBUG(DBG_GENERIC_DRV, 
		   ("starting sendtoServer with socket %d\n", msocket));

	h1->dummy1 = htolel(PLATFORM);
	h1->dummy2 = 0;
	h1->DLLVersion = htolel(DLLVERSION);

	count = sizeof(UDPH1_alt);
	if ((D->ip_version >= 252) && (D->ip_version <260))
			count = sizeof(UDPH1);
	else if (D->ip_version >= 264)
			count = sizeof(UDPH1);

	// wait for writefd to become ready
	FD_ZERO(&rset);
	FD_SET((SOCKET_TYPE)msocket, &rset);
	tmp = select(msocket+1, NULL, &rset, NULL, tv);

        if (tmp > 0) {
		res = sendto(msocket, (char *) h1, count, 0,
			     (struct sockaddr *)server, sizeof(*server));
                if (res == -1)
			// send failed
                        perror("sendtoServer sendto()");
        } else {
		// socket blocked for writing
                perror("sendtoServer select()");
        }

	PRINTDEBUG(DBG_GENERIC_DRV, ("sendtoserver returned %d\n", res));
	return res;
}


int sendtoServer3(SOCKET_TYPE msocket, struct sockaddr_in *server, UDPR3 *r3, struct timeval *tv)
{
	int res = k_UDPtimeout;
	fd_set rset;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, NULL, &rset, NULL, tv) > 0)
		res = sendto(msocket, (char *) r3, sizeof(UDPR3), 0, 
			     (struct sockaddr *) server, 
			     sizeof(struct sockaddr_in));

	return res;
}


int recvfromServer(SOCKET_TYPE msocket,UDPR1 * r1,struct timeval *tv)
{
	int res = -1;
	int count = 0;
	fd_set rset;

	RESET_TIMEOUT(*tv)

nochmal_recvfromServer:;

	r1->status = 0;					// Neu 10.02.04

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
	  	res = recvfrom(msocket, (char *) r1, sizeof(UDPR1), 0, NULL, 0);
		if (res == sizeof(UDPR1))
		{
			if ((unsigned long)(letohl(r1->paket_index)) != 
			    D->udp_paket_count) 
				goto nochmal_recvfromServer;
			res = k_ok;

			if (letohl(r1->status) == k_TryLater)
				res = k_TryLater;
		}
		else
		{
			if (count < k_max_retry_udp	)
			{
				count = count + 1;
				goto nochmal_recvfromServer;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;
	return(res);
}


int recvfromServer2(SOCKET_TYPE msocket, UDPR2 * r2, struct timeval *tv)
{
	int res = -1;
	int count = 0;
	fd_set rset;

	RESET_TIMEOUT(*tv)

nochmal_recvfromServer2:;

	r2->status = 0;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
		res = recvfrom(msocket, (char *) r2, sizeof(UDPR2), 0, NULL, 0);
		if (res == sizeof(UDPR2))
		{
			if ((unsigned long)(letohl(r2->paket_index)) != D->udp_paket_count) 
				goto nochmal_recvfromServer2;
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))
			{
				if (r2->status == letohl(k_TryLater))
				{
					if ((unsigned long)(letohl(r2->paket_index)) == D->udp_paket_count)
					{
						return k_TryLater;
					}
				}
			}
			// TODO: commented out in MGs code
			if (count < k_max_retry_udp)
			{
				count = count + 1;
				goto nochmal_recvfromServer2;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;
	
	return res;
}


int recvfromServer3(SOCKET_TYPE msocket, UDPR3 * r3, struct timeval *tv)
{
	int res = -1;
	int count = 0;
	fd_set rset;

	RESET_TIMEOUT(*tv)

nochmal_recvfromServer3:;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
		res = recvfrom(msocket, (char *) r3, sizeof(UDPR3), 0, NULL, 0);
		if (res == sizeof(UDPR3))
		{
			if ((unsigned long)(letohl(r3->paket_index)) != D->udp_paket_count) 
				goto nochmal_recvfromServer3;
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))
			{
				if ((unsigned long)(letohl(r3->PacketNr)) == D->udp_paket_count) 
				{
					return letohl(r3->paket_index);
				}
			}
			// TODO: commented out in MGs code
			if (count < k_max_retry_udp)
			{
				count = count + 1;
				goto nochmal_recvfromServer3;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;

	return res;
}

int recvfromServer4(SOCKET_TYPE msocket,UDPR4 *r4,struct timeval *tv)
{
	int res = -1;
	int count = 0;
	fd_set rset;

	RESET_TIMEOUT(*tv)

nochmal_recvfromServer4:;

	r4->status = 0;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);
	RESET_TIMEOUT (*tv)

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
		res = recvfrom(msocket, (char *) r4, sizeof(UDPR4), 0, NULL, 0);
		if (res == sizeof(UDPR4))
		{
			if ((unsigned long)(letohl(r4->paket_index))!= D->udp_paket_count)
			{
				goto nochmal_recvfromServer4;
			}
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))
			{
				if (letohl(r4->status) == k_TryLater)
				{
					if ((unsigned long)(letohl(r4->paket_index)) == D->udp_paket_count)
					{
						return k_TryLater;					// gendert am 27.08.03
						//return k_ok;
					}
				}
			}
			// TODO: commented out in MGs code
			if (count < k_max_retry_udp	)
			{
				count = count + 1;
				goto nochmal_recvfromServer4;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;

	return res;
}

int do_udp_command(int sockfd, UDPH1 *h1, UDPR1 *r1, struct sockaddr_in *server, struct timeval *tv)
{
	/* Liefert bei Fehler 1 */
	int res, retry;
	PRINTDEBUG(DBG_GENERIC_DRV, ("do_udp_command sockfd %d\n", sockfd));

	retry = 0;

	udp_retry:;

	RESET_TIMEOUT (*tv);

	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = htolel(D->udp_paket_count);
	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}
	res = recvfromServer(sockfd, r1, tv);
	if (k_ok != res)
	{
		// Fehlerbehandlung
		if (res > 0)
		{
			D->fehlercode = res;
			return(1);
		}
		if (letohl(r1->status) == k_TryLater)
		{
			SLEEP (10);
			goto udp_retry;
		}
		D->fehlercode = k_error_tcp_error;
		if (res == k_UDPtimeout)
		{
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry < k_max_retry_udp)
		{
			retry = retry + 1;
			incRetryCounter();
			goto udp_retry;
		}
		return(1);
	}
	if (letohl(r1->status) == k_DataZuKlein)
	{
		D->fehlercode = k_Data_is_to_small;
		return(1);
	}
	if (letohl(r1->status) == k_WrongPassword)
	{
		D->fehlercode = k_error_tcp_password;
		return(1);
	}
	if (letohl(r1->status) != k_ok)
	{
		D->fehlercode = letohl(r1->status);
		return(1);
	}
	return(0);
}


int do_udp_command2(int sockfd, UDPH1 *h1, UDPR2 *r2, struct sockaddr_in *server, struct timeval *tv)
{
	int res, retry;
	retry = 0;

udp_retry:;
	RESET_TIMEOUT (*tv)

	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = htolel(D->udp_paket_count);

	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}
	res = recvfromServer2(sockfd, r2, tv);
	if (k_ok != res)
	{
		// Fehlerbehandlung
		if (res > 0)
		{
			D->fehlercode = res;
			return(1);
		}
		if (letohl(r2->status) == k_TryLater)
		{
			SLEEP (10);
			goto udp_retry;
		}
		D->fehlercode = k_error_tcp_error;
		if (res == k_UDPtimeout)
		{
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry  < k_max_retry_udp)
		{
			retry = retry + 1;
			incRetryCounter();
			goto udp_retry;
		}
		return(1);
	}
	if (letohl(r2->status) == k_WrongPassword)
	{
		D->fehlercode = k_error_tcp_password;
		return(1);
	}
	if (letohl(r2->status) != k_ok)
	{
		D->fehlercode = letohl(r2->status);
		return(1);
	}
	D->fehlercode = 0;
	return (0);
}

int do_udp_command_nx4(SOCKET_TYPE sockfd, UDPH1 *h1, UDPR4 *r4, struct sockaddr_in *server, struct timeval *tv,
					   int blockcount, char *buffer, short *datatyp)
{
	int res,retry,lauf;

	retry = 0;
udp_retry:;
	RESET_TIMEOUT (*tv)

	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = htolel(D->udp_paket_count);

	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}

	for (lauf = 0; lauf < blockcount; lauf++)
	{
		res = recvfromServer4(sockfd,r4,tv);
		if (k_ok != res)
		{
			// Fehlerbehandlung

			if (res > 0)
			{
				D->fehlercode = res;
				return(lauf);
			}

			if (letohl(r4->status) == k_TryLater)
			{
				SLEEP (10);
				goto udp_retry;
			}

			D->fehlercode = k_error_tcp_error;

			if (res == k_UDPtimeout)
			{
				D->fehlercode = k_error_tcp_timout_read;
			}
			if (retry < k_max_retry_udp)
			{
				retry = retry + 1;
				incRetryCounter();
				goto udp_retry;
			}
			return(lauf);
		}
		if (letohl(r4->status) == k_WrongPassword)
		{
			D->fehlercode = k_error_tcp_password;
			return(lauf);
		}

		if (letohl(r4->status) != k_ok)
		{
			D->fehlercode = letohl(r4->status);
			return(lauf);
		}
		if (lauf == 0)
			datatyp[0] = (short)(letohl(r4->datatyp));
		if (letohl(r4->PacketNr) != lauf)
		{
			D->fehlercode = k_error_tcp_timout_read;
			return lauf;
		}
		memcpy_letohl(&buffer[lauf * 350 * 4],r4->value,350 * 4);
	}

#ifndef _MYUDPTEST
	D->fehlercode = 0;
#endif //_MYUDPTEST
	return (blockcount);
}

void init_packet(struct timeval *time, UDPH1 *h1) {
	time->tv_sec = D->timeout / 1000; 
	time->tv_usec = (D->timeout % 1000) * 1000;

	h1->timeout = htolel(D->timeout);
	h1->link = htolel(D->hostlink);
	strcpy(h1->password, D->passwd);
}

/**************************************************************************
 * the real ADwin functions start here
 **************************************************************************/

void NAME (Boot) (const char *dateiname) {
	/* Returnwert <= 1000 => Fehler */

	int res,retry,filesize;
	FILE *fd;
	int32_t pos_in_name,lauf;

	UDPH1 h1;			
	UDPR1 r1;			
	struct timeval my_time;  

	short help;
	unsigned char *daten;

#ifndef ADWIN_EMBEDDED
	lauf = strlen(dateiname);
	if ((lauf < 4)
	    || (dateiname[lauf-4] != '.') 
	    || ((dateiname[lauf-3] != 'B') && (dateiname[lauf-3] != 'b'))
	    || ((dateiname[lauf-2] != 'T') && (dateiname[lauf-2] != 't'))
	    || ((dateiname[lauf-1] != 'L') && (dateiname[lauf-1] != 'l')))
	{
		//keine MessageBox!!
		D->fehlercode = k_File_not_btl; //MS changed from k_File_not_binaer; //
		return;// 0;
	}
	
	
	help = 0;
	
//    if (D->hostlink <= 0) return;//(1);
	
	//Test if the filename is a valid name.
	//It tests only if there is a '9', a '10' or an '11' bevor the point(.)!!!	
	pos_in_name = lauf - 5;
	if (pos_in_name < 1)
		pos_in_name = 1;
	
	if ((dateiname[pos_in_name + 1] != '.') 
	    || ( ! (dateiname[pos_in_name] == '9'
		    || ((dateiname[pos_in_name - 1] == '1') && (dateiname[pos_in_name] == '0'))
		    || ((dateiname[pos_in_name - 1] == '1') && (dateiname[pos_in_name] == '1'))))) {
		
		D->fehlercode = k_File_not_valid;
		PRINTDEBUG(DBG_GENERIC_DRV, ("Boot failed - file not valid\n"));
		return;
	}
#endif
	retry = 0;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_Boot);
	h1.value[0] = htolel(1000); // Iboardsize
	
	if ((fd = fopen (dateiname, "rb")) == NULL) 
	{
		//keine MessageBox!!
		D->fehlercode = k_File_not_found;//k_wrong_version;

		return;//(1); //0 bei M.G.
	}

	filesize= get_filesize(fd);
	h1.value[1] = htolel(5 << 16);		// PaketCount
	h1.value[2] = htolel(k_proc_t9);
	h1.value[3] = htolel(filesize);

	daten = (unsigned char *)malloc(filesize);
	fread(daten, sizeof(char), filesize, fd);
	fclose(fd);

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
	{
		free(daten);
		return;//(1);
	}

	if (setarray(daten, filesize, sizeof(char), D->sockfd, 
		     (struct sockaddr *) &(D->server), 5, &my_time, 0))
	{
		free(daten);
		return;//(1);  //255 bei M.G.
	}
	free(daten);
 udp_retry:;
	res = recvfromServer(D->sockfd, &r1, &my_time);

	if (k_ok != res)
	{
		// Fehlerbehandlung
		
		if (res > 0)
		{
			D->fehlercode = res;
			return;//(1);   //0 bei M.G.
		}
		
		D->fehlercode = k_error_tcp_error;
		
		if (res == k_UDPtimeout)
		{
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry < 5)
		{
			retry = retry + 1;
			incRetryCounter();
			goto udp_retry;
		}
		return;//(1);//0 bei M.G.
	}
	
	if (letohl(r1.status) != k_ok)
	{
		//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
		switch (letohl(r1.status))
		{
		case 1:
			D->fehlercode = k_timeout_to_link;
			break;
		case 2:
			D->fehlercode = k_timeout_from_link;
			break;
		case 5:
			D->fehlercode = k_timeout_to_link;
			break;
		case -1:
		case -10:
		case -15:
		case -20:
		case -25:
		case -26:
		case -30:
		case -33:
		case -40:
		case -41:
			D->fehlercode = k_error_tcp_error;
			break;
		case -5:
			D->fehlercode = k_error_tcp_password;
			break;
			
		case -31:
			D->fehlercode = k_Data_is_to_small;
			break;
			
		case -32:
			D->fehlercode = k_Fifo_to_less_data;
			break;
		}
		return;//(1); //0 bei M.G.
	}
	else
		D->fehlercode = k_ok;
	
	
	PRINTDEBUG(DBG_GENERIC_DRV, ("Boot returned successfully\n"));
	
	return;
}

int32_t NAME (Test_Version) () {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);
	h1.befehl   = htolel(k_TestVersion);

	if (do_udp_command(D->sockfd, &h1,&r1,&(D->server),&my_time )) return 3;
//	Prozessor= r1.value[1];
//	Kartentyp= r1.value[2];
//	Treiberversion = r1.value[3];

#ifndef _MYUDPTEST
        D->fehlercode = 0;
#endif
	return 0;
}

processor_type_t NAME (Processor_Type) ()
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_TestVersion);

	if (do_udp_command(D->sockfd, &h1,&r1, &(D->server), &my_time )) return 0;

#ifndef _MYUDPTEST
	D->fehlercode = 0;
#endif
	switch (letohl(r1.value[1])) {
	case k_proc_net_t9:  return PROCESSOR_T9;
	case k_proc_net_t10: return PROCESSOR_T10;
	case k_proc_net_t11: return PROCESSOR_T11;
	default:             return (letohl(r1.value[1]));
	}
}


int32_t NAME (Workload) (int32_t Priority) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;
	unsigned short wert;
	int32_t wert1,zeit1;
	float help;

	init_packet(&my_time, &h1);

	h1.befehl = htolel(k_Auslastung);
	h1.value[0] = htolel(k_Auslastung);
	h1.value[1] = 0;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(100); //255 bei M.G.

	if (letohl(r1.value[2]) == 0)
	{
		wert1 = letohl(r1.value[0]);
		zeit1 = letohl(r1.value[1]);
		if (zeit1 == 0) wert = 100;
		else
		{
			processor_type_t processor;
			processor = NAME (Processor_Type) ();
			if (processor < 0)
				return 255;

			help = (float)wert1 / (float)zeit1;
			if (processor == PROCESSOR_T11) 
				wert = (unsigned short)(100. * (((1./9.) - help) * 9));
			else 
				wert = (unsigned short)(100. * ((0.25 - help) / 0.25));
		}
	}
	else
		wert = (unsigned short)(letohl(r1.value[0]));

	if (wert > 100) 
		wert = 100;

	return(wert);
}

int32_t NAME (Free_Mem) (mem_spec_t mem_spec) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_3plus1);
 	h1.value[0] = htolel(k_MemoryAll);
 	h1.value[1] = htolel((int32_t) mem_spec);
 	h1.value[2] = 0;

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) 
		return(1);

#ifndef BUILD_LIBRARY
	processor_type_t processor;
	
	/* first, we need to determine the processor type */
	processor = NAME (Processor_Type) ();
	/* todo: if this fails, we have two timeouts here... */

	switch (mem_spec) {
	case MEM_PM_LOCAL:
		if ((processor == PROCESSOR_T9) || (processor == PROCESSOR_T10))
			return letohl(r1.value[0]) * 6;
		else 
			return letohl(r1.value[0]) * 4;
		break;
	case MEM_ALL:
	case MEM_EM_LOCAL:
	case MEM_DM_LOCAL:
	case MEM_DRAM_EXTERN:
		return letohl(r1.value[0]) * 4;
	default: 
		return 0;
	}
#else
	// conversion will be done in libadwin.c
	return letohl(r1.value[0]);
#endif
}

void NAME (Load_Process) (const char *dateiname)
{
 	int res,filesize,retry;
	int lauf;
 	FILE *fd;
 	unsigned char *daten;

	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

#ifndef ADWIN_EMBEDDED
	lauf = strlen(dateiname);

	if ((dateiname[lauf-4] != '.') || ((dateiname[lauf-3] != 'T') && (dateiname[lauf-3] != 't')))
	{
		D->fehlercode = k_File_not_binaer;
		return;// 0;
	}
#endif
 	retry = 0;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_LadeBin);

 	if ((fd = fopen (dateiname, "rb")) == NULL)
 	{
 		D->fehlercode = k_File_not_found;
 		return;
 	}

 	filesize= get_filesize(fd);
 	daten = (unsigned char *)malloc(filesize);
 	fread(daten, sizeof(char), filesize, fd);
 	fclose(fd);

 	h1.value[1] = htolel(5 << 16);	 // PaketCount
 	h1.value[2] = htolel(k_proc_t9);
 	h1.value[3] = htolel(filesize);

	// TODO: replace returns with "goto: errcode-mapping"

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
	{
		free(daten);
		return;
	}

 	if (setarray(daten, filesize, sizeof(char), D->sockfd,(struct sockaddr *) &(D->server), 5, &my_time, 0))
 	{
 		free(daten);
 		return;
 	}
 	free(daten);

udp_retry:;

 	res = recvfromServer(D->sockfd, &r1, &my_time);
 	if (k_ok != res)
 	{
 		// Fehlerbehandlung

 		if (res > 0)
 		{
 			D->fehlercode = res;
 			return;
 		}

 		D->fehlercode = k_error_tcp_error;

 		if (res == k_UDPtimeout)
 		{
 			D->fehlercode = k_error_tcp_timout_read;
 		}
 		if (retry < 5)
 		{
 			retry = retry + 1;
			incRetryCounter();
 			goto udp_retry;
 		}
 		return;
 	}

	// todo errcode_mapping:;

	//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
	switch (letohl(r1.status))
	{
	case k_ok: D->fehlercode = 0;
		break;
	case 1:
		D->fehlercode = k_timeout_to_link;
		break;
	case 2:
		D->fehlercode = k_timeout_from_link;
		break;
	case 5:
		D->fehlercode = k_timeout_to_link;
		break;
	case -30:
		D->fehlercode = k_wrong_version;
		break;
	case -1:
	case -10:
	case -15:
	case -20:
	case -25:
	case -26:
	case -33:
	case -40:
	case -41:
		D->fehlercode = k_error_tcp_error;
		break;
	case -5:
		D->fehlercode = k_error_tcp_password;
		break;
		
	case -31:
		D->fehlercode = k_Data_is_to_small;
		break;
		
	case -32:
		D->fehlercode = k_Fifo_to_less_data;
		break;
	case -35:
		D->fehlercode = k_WrongBinFile;
		break;
	}
}

void NAME (Start_Process) (int32_t ProcessNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_3plus1);
 	h1.value[0] = htolel(k_Start);
 	h1.value[1] = htolel(ProcessNo);
 	h1.value[2] = 0;
 	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}


void NAME (Stop_Process) (int32_t ProcessNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_3plus1);
 	h1.value[0] = htolel(k_Stop);
 	h1.value[1] = htolel(ProcessNo);
 	h1.value[2] = 0;
 	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}


void NAME (Clear_Process) (int32_t ProcessNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_3plus1);
 	h1.value[0] = htolel(k_ClearProcess);
 	h1.value[1] = htolel(ProcessNo);
 	h1.value[2] = 0;

 	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}

int32_t NAME (Process_Status) (int32_t process_no) {
	return NAME (Get_Par) (-100 + process_no);
}

void    NAME (Set_ProcessDelay) (int32_t process_no, int32_t value) {
	NAME (Set_Par) (-90 + process_no, value);
}

int32_t NAME (Get_ProcessDelay) (int32_t process_no) {
	return NAME (Get_Par) (-90 + process_no);
}

void NAME (Set_Par) (int32_t Index, int32_t Value) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_SetPar);
	h1.value[1] = htolel(1000 + Index);
	h1.value[2] = htolel(Value);

	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}

int32_t NAME (Get_Par) (int32_t Index) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_GetPar);
	h1.value[1] = htolel(1000 + Index);
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) 
		return(1);

	return letohl(r1.value[0]);
}

void NAME (Get_Par_Block) (int32_t *pData, int32_t StartIndex, int32_t Count)
{
	UDPH1 h1;
	UDPR2 r2;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_GetParAll);
	h1.value[0] = htolel(StartIndex);
	h1.value[1] = htolel(Count);

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;//1;
	memcpy_letohl(pData, r2.value, Count * sizeof(int32_t));
}

void NAME (Get_Par_All) (int32_t *pData)
{
	UDPH1 h1;
	UDPR2 r2;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_GetParAll);
	h1.value[0] = htolel(1); //np;
	h1.value[1] = htolel(80);//anzahl;
	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;//1;

	memcpy_letohl(pData, r2.value, 80 * sizeof(int32_t));
}


void NAME (Set_FPar) (int32_t Index, float Value)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;
	int32_t *lptr;

	init_packet(&my_time, &h1);

	lptr = (void*)&Value;

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_SetPar);
	h1.value[1] = htolel(1100 + Index);
	h1.value[2] = htolel(*lptr);

	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}


float NAME (Get_FPar) (int32_t Index)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_GetPar);
	h1.value[1] = htolel(1100 + Index);
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
		return(1);
	
	// interpret int in structure as float
	return *(float *)((void *)&r1.value[0]);
}

void NAME (Get_FPar_Block) (float *pData, int32_t StarIndex, int32_t Count) {

	UDPH1 h1;
	UDPR2 r2;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_GetParAll);
	h1.value[0] = htolel(100 + StarIndex);
	h1.value[1] = htolel(Count);

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;// 1;

	memcpy_letohl(pData, r2.value, Count * sizeof(int32_t));
}

void NAME (Get_FPar_All) (float *pData) {

	UDPH1 h1;
	UDPR2 r2;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_GetParAll);
	h1.value[0] = htolel(100 + 1);
	h1.value[1] = htolel(80);

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;// 1;

	memcpy_letohl(pData, r2.value, 80 * sizeof(int32_t));
}

int32_t NAME (Data_Length) (int32_t DataNo)
{
	int lh;
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_GetDataLength);
	h1.value[1] = htolel(DataNo);

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time ))
	{
		return 0;														// gendert 08.05.02
	}
	lh = letohl(r1.value[0]);
	if (lh > 0) lh = lh - 1;

	return lh;
}

void NAME (SetData_Long) (int32_t DataNo, int32_t *Data, int32_t Startindex, int32_t Count) {
	NAME (Set_Data) (DataNo, Data, ADWIN_LONG, Startindex, Count);
}

void NAME (GetData_Long) (int32_t data_no, int32_t *dest, int32_t start_index, int32_t count) {
	NAME (Get_Data) (data_no, dest, ADWIN_LONG, start_index, count);
}

void NAME (SetData_Float) (int32_t DataNo, float *Data, int32_t Startindex, int32_t Count) {
	NAME (Set_Data) (DataNo, Data, ADWIN_FLOAT, Startindex, Count);
}

void NAME (GetData_Float) (int32_t DataNo, float *Data, int32_t Startindex, int32_t Count) {
	NAME (Get_Data) (DataNo, Data, ADWIN_FLOAT, Startindex, Count);
}
int32_t NAME (Fifo_Empty) (int32_t FifoNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_GetFifoEmpty);
	h1.value[1] = htolel(FifoNo);
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(255);
	return letohl(r1.value[0]);
}
int32_t NAME (Fifo_Full) (int32_t FifoNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_GetFifoCount);
	h1.value[1] = htolel(FifoNo);
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(255);

	return letohl(r1.value[0]);
}



void NAME (Fifo_Clear) (int32_t FifoNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_3plus1);
 	h1.value[0] = htolel(k_ClearFifo);
 	h1.value[1] = htolel(FifoNo);
 	h1.value[2] = 0;
 	do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time);
}
void NAME (SetFifo_Long) (int32_t FifoNo, int32_t *Data, int32_t Count) {
	NAME (Set_Fifo) (FifoNo, 2, Data, Count);
}

void NAME (SetFifo_Float) (int32_t FifoNo, float *Data, int32_t Count) {
	NAME (Set_Fifo) (FifoNo, 5, Data, Count);
}

void NAME (GetFifo_Long) (int32_t FifoNo, int32_t *Data, int32_t Count) {
	NAME (Get_Fifo) (FifoNo, 2, Data, Count);
}

void NAME (GetFifo_Float) (int32_t FifoNo, float *Data, int32_t Count) {
	NAME (Get_Fifo) (FifoNo, 5, Data, Count);
}

void    NAME (Data2File) (const char *filename, int32_t data_no, 
			  int32_t start_index, int32_t count, int mode) {
	FILE *file;
	void *data;

	file = fopen (filename, mode ? "ab" : "wb");
	if (file == NULL)
	{
		D->fehlercode = 200;
		return;
	}
	data = malloc (count * sizeof (int32_t));
	NAME (Get_Data) (data_no, data, (short)(NAME (GetDataType)) ((short)data_no), start_index, count);
	fwrite (data, 1, count, file);
	free (data);
	fclose (file);
}

void NAME (SetData_String) (int32_t data_no, const char *src)
{
	uint32_t *values;
	int32_t lauf,anzahl;


	anzahl = strlen (src);
	values = (uint32_t*)malloc((anzahl + 2) * sizeof (int32_t));

	values[0] = anzahl;

	for (lauf = 0; lauf < anzahl; lauf ++)
		values[lauf + 1] = (unsigned char) src[lauf];

	values[lauf + 1] = 0;

	NAME (Set_Data) (data_no, values, 2, 1, anzahl + 2);

	free (values);
}

int32_t NAME (GetData_String) (int32_t DataNo, char *Data, int32_t MaxCount)
{
	int32_t *values;
	int32_t lauf, res;

	values = (int32_t *) malloc ((MaxCount + 2) * sizeof(int32_t));

	NAME (Get_Data) (DataNo, values, 2, 1, MaxCount + 1);
	if (D->fehlercode != 0)
	{
		// ging nicht
		free (values);
		return -1;
	}
	res = values[0];
	if (MaxCount - 1 < res) res = MaxCount - 1;

	for (lauf = 0; lauf < res; lauf ++)
	{
		Data[lauf] = (char) values[lauf + 1];
	}

	Data[lauf] = 0;

	free (values);
	return res;
}
int32_t NAME (String_Length) (int32_t DataNo)
{
	int32_t value;

	NAME (Get_Data) (DataNo, &value, 2, 1, 1);
	if (D->fehlercode != 0)
		return -1;

	return value;
}


int32_t NAME (Get_Last_Error) ()
{
	int32_t res;
	res = D->fehlercode;
	D->fehlercode = 0;
	return res;
}

int32_t NAME (Get_Last_Error_Text) (int32_t errorcode, char *text, int32_t textlaenge)
{
	int count;
	char Buffer[200];

	switch(errorcode)
	{
		case 0:
			if (english)
				strcpy(Buffer,k_error_0_e);
			else
				strcpy(Buffer,k_error_0_d);
		break;
		/*case 1:
			if (english)
				strcpy(Buffer,k_error_1_e);
			else
				strcpy(Buffer,k_error_1_d);
		break;

		case 2:
			if (english)
				strcpy(Buffer,k_error_2_e);
			else
				strcpy(Buffer,k_error_2_d);
		break;

		case 3:
			if (english)
				strcpy(Buffer,k_error_3_e);
			else
				strcpy(Buffer,k_error_3_d);
		break;

		case 4:
			if (english)
				strcpy(Buffer,k_error_4_e);
			else
				strcpy(Buffer,k_error_4_d);
		break;

		case 10:
			if (english)
				strcpy(Buffer,k_error_10_e);
			else
				strcpy(Buffer,k_error_10_d);
		break;

		case 11:
			if (english)
				strcpy(Buffer,k_error_11_e);
			else
				strcpy(Buffer,k_error_11_d);
		break;

		case 15:
			if (english)
				strcpy(Buffer,k_error_15_e);
			else
				strcpy(Buffer,k_error_15_d);
		break;*/

		case 20:
			if (english)
				strcpy(Buffer,k_error_20_e);
			else
				strcpy(Buffer,k_error_20_d);
		break;

		case 100:
			if (english)
				strcpy(Buffer,k_error_100_e);
			else
				strcpy(Buffer,k_error_100_d);
		break;

		/*case 101:
			if (english)
				strcpy(Buffer,k_error_101_e);
			else
				strcpy(Buffer,k_error_101_d);
		break;*/

		case 102:
			if (english)
				strcpy(Buffer,k_error_102_e);
			else
				strcpy(Buffer,k_error_102_d);
		break;

		case 200:
			if (english)
				strcpy(Buffer,k_error_200_e);
			else
				strcpy(Buffer,k_error_200_d);
		break;

		/*case 201:
			if (english)
				strcpy(Buffer,k_error_201_e);
			else
				strcpy(Buffer,k_error_201_d);
		break;*/

		case 202:
			if (english)
				strcpy(Buffer,k_error_202_e);
			else
				strcpy(Buffer,k_error_202_d);
		break;

		case 203:
			if (english)
				strcpy(Buffer,k_error_203_e);
			else
				strcpy(Buffer,k_error_203_d);
		break;

		/*case 204:
			if (english)
				strcpy(Buffer,k_error_204_e);
			else
				strcpy(Buffer,k_error_204_d);
		break;*/
		case 205:
			if (english)
				strcpy(Buffer,k_error_205_e);
			else
				strcpy(Buffer,k_error_205_d);
		break;

		
		/*
		case 1000:
			if (english)
				strcpy(Buffer,k_error_1000_e);
			else
				strcpy(Buffer,k_error_1000_d);
		break;

		case 1001:
			if (english)
				strcpy(Buffer,k_error_1001_e);
			else
				strcpy(Buffer,k_error_1001_d);
		break;

		case 1002:
			if (english)
				strcpy(Buffer,k_error_1002_e);
			else
				strcpy(Buffer,k_error_1002_d);
		break;*/

		/*case 1003:
			if (english)
				strcpy(Buffer,k_error_1003_e);
			else
				strcpy(Buffer,k_error_1003_d);
		break;*/

		case 2000:
			if (english)
				strcpy(Buffer,k_error_2000_e);
			else
				strcpy(Buffer,k_error_2000_d);
		break;

		case 2001:
			if (english)
				strcpy(Buffer,k_error_2001_e);
			else
				strcpy(Buffer,k_error_2001_d);
		break;

		case 2002:
			if (english)
				strcpy(Buffer,k_error_2002_e);
			else
				strcpy(Buffer,k_error_2002_d);
		break;

		/*case 3000:
			if (english)
				strcpy(Buffer,k_error_3000_e);
			else
				strcpy(Buffer,k_error_3000_d);
		break;

		case 3001:
			if (english)
				strcpy(Buffer,k_error_3001_e);
			else
				strcpy(Buffer,k_error_3001_d);
		break;*/

		default:
			if (english)
				strcpy(Buffer,k_error_un_e);
			else
				strcpy(Buffer,k_error_un_d);
		break;

	}
	count = strlen(Buffer) + 1;
	if (count > textlaenge) count = textlaenge;

	textlaenge = count;

	memcpy(text, Buffer, textlaenge);

	text[textlaenge - 1] = 0;

	return(textlaenge - 1);
}

/////////////////////////////////////////// undocumented: todo


/* Well, what should this function return or do on error? It is tricky: */

/* On communication error, we return -1. That's simple. */

/* If the data on the adwin system was not defined , this function
   cannot set the error code properly, because different error codes
   exist for fifos and datas. However, this function has noe clue
   whether it asks for the data type of a fifo or a data. So in that
   case it returns 0, and the calling function must make sure that the
   error code is properly set. */
adwin_system_data_type_t NAME (GetDataType) (short data_no) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	// 0 Fehler oder nicht deklariert, 3 long, 5 float
	h1.befehl   = htolel(k_GetDataType);
	h1.value[0] = htolel(k_GetData);
	h1.value[1] = htolel(data_no);
	h1.value[2] = htolel(1);
	h1.value[3] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) 
		return(-1);
	
	adwin_system_data_type_t tmp = letohl(r1.value[0]);
	tmp = check_system_type(tmp);
	
	if (tmp == ADWIN_SYSTEM_INTEGER)
		tmp = ADWIN_SYSTEM_LONG;

	return tmp;
}


/***************Fifos**********************************************************************************/

short Get_Fifo_UDP (char *Data, short FifoNo, int32_t Count, short dsp_typ, 
		    int32_t *pkt_index, short flag_retry, struct sockaddr_in *server)
{
	UDPH1 h1;
	UDPR3 r3;
	struct timeval my_time;
	int retry;
	int size, pointer, zaehler;

Get_Fifo_UDP_retry:														// Neu 10.02.04
	init_packet(&my_time, &h1);

	D->fehlercode = 0;

	if (flag_retry)
		h1.befehl   = htolel(k_GetFifoRetry);
	else
		h1.befehl   = htolel(k_GetFifo);


	h1.value[0] = htolel(dsp_typ);
	h1.value[1] = htolel(FifoNo | (5 << 16));		// ?
	if (flag_retry)
		h1.value[2] = htolel(*pkt_index);

	h1.value[3] = htolel(Count);

	size = Count * sizeof(int32_t);

	retry = 0;

//udp_retry:;	 //bei M.G. enthalten, wird aber nicht aufgerufen

	D->udp_paket_count = D->udp_paket_count + 1;
	h1.paket_index = htolel(D->udp_paket_count);

	if (!flag_retry)
		*pkt_index = D->udp_paket_count;

	// senden Standardpaket
	if (k_UDPtimeout == sendtoServer(D->sockfd, server, &h1, &my_time))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(-1);
	}

	// Nun max. 9800 Byte entspricht 7 Pakete
	pointer = 0;
	zaehler = 0;
	while(size > 0)
	{
		int res;

		res = recvfromServer3(D->sockfd, &r3, &my_time);
		if (res == k_ok)
		{
			// Koennte noch Fehlercode sein

			if (letohl(r3.PacketNr) == zaehler)
			{
				if (size >= sizeof(r3.value))
					memcpy_letohl(&Data[pointer],r3.value,sizeof(r3.value));
				else
					memcpy_letohl(&Data[pointer],r3.value,size);//todo
				size = size - sizeof(r3.value);
				pointer = pointer + sizeof(r3.value);
				zaehler = zaehler + 1;
			}
		}
		else
		{
			if (res == k_TryLater)							// Neu 10.02.04
			{												// Neu 10.02.04
				SLEEP (10);									// Neu 10.02.04
				goto Get_Fifo_UDP_retry;					// Neu 10.02.04
			}												// Neu 10.02.04
			if ((res == k_RetryUnknown) && flag_retry)
				return 200;
			if (res == k_UDPtimeout)
			{
				return 255;
			}
			if (res == k_FifoLessData)
				D->fehlercode = k_Fifo_to_less_data;
			return 150;
			}
		}
	return 0;
}

#define GET_FIFO_BLOCK_SIZE 6


// if used as generic_driver, type must be adwin_data_type. if used
// with libadwin.c, type must be adwin_system_data_type, so that no
// conversion happens
void    NAME (Get_Fifo) (int32_t fifo_no, adwin_data_type_t type, void *dest, int32_t count) {
	short dsp_typ;
	char *daten;
	int32_t pkt_index;
	int32_t pkt_pointer,msize;
	short res;

	if (count < 1) return;

	res = 0;
	D->fehlercode = 0;

#ifndef BUILD_LIBRARY
	dsp_typ = (short)NAME (GetDataType) ((short)fifo_no);

	if (dsp_typ == 0) {
		if (D->fehlercode == 0) D->fehlercode = k_Fifo_is_to_small;
		return;
	} 

	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;

	daten = Get_Data_prepare_convert(dsp_typ, dest, type, count);
#else
	dsp_typ = type;
	daten = dest;
#endif	


	msize = count;

	pkt_pointer = 0;

	while (msize > 0) {
		int counter;
		
		counter = 0;
		
	nochmal_von_vorne:
		
		if (msize >= GET_FIFO_BLOCK_SIZE * 350)
			res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, GET_FIFO_BLOCK_SIZE * 350, dsp_typ, &pkt_index, 0, &(D->server));
		else
			res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, msize, dsp_typ, &pkt_index ,0, &(D->server));
		
		while ((res != 0) && (counter < 5)) {
			switch (res) {
			case 255: {
				if (msize >= GET_FIFO_BLOCK_SIZE * 350)  //9800 / sizeof(long)
					res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, GET_FIFO_BLOCK_SIZE * 350, dsp_typ,  &pkt_index, 1, &(D->server));
				else
					res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, msize, dsp_typ, &pkt_index ,1, &(D->server));
			}
				break;
				
			case 200: goto nochmal_von_vorne;
				break;
			case 150:
				goto free_and_return;
			case 100:
				D->fehlercode = k_Fifo_to_less_data;
				goto free_and_return;
			case -1:
				D->fehlercode = k_error_tcp_timout_read;
				goto free_and_return;
			}
			counter = counter + 1;
			
			if (counter >= 5) {
				D->fehlercode = k_error_tcp_timout_read;
				goto free_and_return;
			}
		}
		msize = msize - GET_FIFO_BLOCK_SIZE * 350;
		pkt_pointer = pkt_pointer + GET_FIFO_BLOCK_SIZE * 350 * sizeof(int32_t);
	}
	
	D->fehlercode = 0;
	
#ifndef BUILD_LIBRARY
	Get_Data_do_convert(dsp_typ, daten, type, Data, count);
#endif
	
 free_and_return:

	if (dest != daten)
		free(daten);
}


short Set_Fifo_UDP (char *Data, short FifoNo, int32_t Count ,short dsp_typ, int32_t *pkt_index ,short flag_retry, struct sockaddr_in *server)
{
	int size, pointer, zaehler;
	UDPH1 h1;
	UDPR1 r1;
	UDPR3 r3;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	D->fehlercode = 0;

	if (flag_retry)
		h1.befehl   = htolel(k_SetFifoRetry);
	else
		h1.befehl   = htolel(k_SetFifo);

	h1.value[0] = htolel(dsp_typ);
	h1.value[1] = htolel(FifoNo | (5 << 16));
	if (flag_retry)
		h1.value[2] = htolel(*pkt_index);
	h1.value[3] = htolel(Count);

	if (do_udp_command(D->sockfd, &h1, &r1, server, &my_time))
	{
		if ((flag_retry) && (letohl(r1.status) == k_SchonBearbeitet))
		{
			D->fehlercode = k_ok;
			return 0;
		}
		else
			return 255;
	}

	if (!flag_retry)
		*pkt_index = D->udp_paket_count;

	// Nun max. 9800 Byte entspricht 7 Pakete
	pointer = 0;
	zaehler = 0;
	size = Count * sizeof(int32_t);

	while(size > 0)
	{
		r3.paket_index = htolel(D->udp_paket_count);
		r3.PacketNr = htolel(zaehler);
		if (sizeof(r3.value) <=  size)
			memcpy_htolel(r3.value,&Data[pointer],sizeof(r3.value));
		else
			memcpy_htolel(r3.value,&Data[pointer],size);

		sendtoServer3(D->sockfd, server, &r3, &my_time);

		size = size - sizeof(r3.value);
		pointer = pointer + sizeof(r3.value);
		zaehler = zaehler + 1;
	}

	// Es fehlt noch letzte Handshake;

	if (recvfromServer(D->sockfd, &r1, &my_time) == k_ok)
	{
		if (letohl(r1.status) == k_ok) return 0;
		if (letohl(r1.status) == k_SchonBearbeitet) return 0;
		if (letohl(r1.status) == k_timeout)
		{
			return 100;
		}
		if (letohl(r1.status) == k_FalschePaketreihenfolge) return 100;
		return 200;
	}
	else
		return 100;
}


void    NAME (Set_Fifo) (int32_t fifo_no, adwin_data_type_t type, void *src, int32_t count) {
	short res;
	int retry;
	short dsp_typ;
	int msize,pkt_pointer;
	int32_t pkt_index;
	char *daten;
	int counter;

	if (count < 1) return;

	D->fehlercode = 0;

#ifndef BUILD_LIBRARY
	dsp_typ = (short)NAME (GetDataType) ((short)fifo_no);

	if (dsp_typ == 0) {
		if (D->fehlercode == 0) D->fehlercode = k_Fifo_is_to_small;
		return;
	}

	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;

	daten = Set_Data_convert(type, src, dsp_typ, count);
#else
	dsp_typ = type;
	daten = src;
#endif

	retry = 0;
	msize = count;
	pkt_pointer = 0;
	while (msize > 0)
	{
		counter = 0;

		if (msize >= 2450)  //9800 / sizeof(long)
			res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, 2450, dsp_typ, &pkt_index, 0, &(D->server));
		else
			res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, msize, dsp_typ, &pkt_index, 0, &(D->server));

		while ((res != 0) && (counter < 5))
		{
			switch (res)
			{
				case 100:
					{
						if (msize >= 2450)  //9800 / sizeof(long)
							res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, 2450, dsp_typ, &pkt_index, 1, &(D->server));
						else
							res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)fifo_no, msize, dsp_typ, &pkt_index, 1, &(D->server));
					}
					break;

				case 200:
					D->fehlercode = k_Fifo_is_to_small;
					goto free_and_return;
				case 255:
					goto free_and_return;
			}
			counter = counter + 1;

			if (counter >= 5)
			{
				D->fehlercode = k_error_tcp_timout_read;
				goto free_and_return;
			}
		}
		msize = msize - 2450;
		pkt_pointer = pkt_pointer + 2450 * sizeof(int32_t);
	}

	D->fehlercode = 0;

 free_and_return:
	if (daten != src)
		free(daten);
	
}
/****Datas***********************************************************************************************/

// 1 Fehler, 0 ok

short getarray (char *arr, int32_t msize, int32_t typsize,SOCKET_TYPE msocket, struct sockaddr *m_serv_addr, int max_paket_count, struct timeval *tv)
{
	int32_t size,last_size,retry,count;
	int32_t arraypointer;
	int32_t anzahl_empfangen;
	int32_t last_arraypointer;
	int32_t packet_nr;
	int fehlerblock;
	UDPR3 message_out3;// this packet is received
	UDPR1 message_out; // this packet is sent
	int value1,res,retry2/*,retry3*/;
	int32_t pNr;
	int paket_erreicht[50];
	int dummy_zaehler;
	int first_empfang;

	int counti;


	fd_set  my_fd_setl;

	counti = 0;
	first_empfang = 0;

	dummy_zaehler = 0;
	anzahl_empfangen = 0;
	fehlerblock = 0;
	arraypointer = 0;
	packet_nr = 1;
	pNr = 1;
	retry = 0;
	retry2 = 0;
	size = msize * typsize; // Gesamtgroesse Datei

	message_out.value[0] = htolel(max_paket_count);

	while(size > 0)
	{
		last_arraypointer = arraypointer;
		last_size = size;

		for (value1 = 0; value1 < max_paket_count; value1++)
			paket_erreicht[value1] = 0;


		for (value1 = 0; value1 < max_paket_count; value1++)
		{
			if (size > 0)
			{
nochmal_empfangen2:;

				FD_ZERO(&my_fd_setl);
				FD_SET(msocket,&my_fd_setl);
				RESET_TIMEOUT (*tv)

				if (1 == select(msocket+1, &my_fd_setl, 0, 0 ,tv))
				{
					struct sockaddr dummy;
					socklen_t len_dummy;

					len_dummy = sizeof(dummy);
					count = recvfrom(msocket, (char *)&message_out3, sizeof(message_out3), 0, &dummy, &len_dummy);
					if (0 < count)
					{
						{
							if (count == sizeof(message_out3))
							{
								if(message_out3.paket_index != letohl(pNr))
								{
									retry = retry + 1;
									incRetryCounter();
									if (retry > k_max_retry_get)
									{
										D->fehlercode = k_error_tcp_error;
										return(1);
									}
									goto nochmal_empfangen2;
								}
								else
								{
									first_empfang = 1;
									anzahl_empfangen = anzahl_empfangen + 1;
									//if (size > sizeof(message_out3.value))
									memcpy_letohl(&arr[arraypointer + letohl(message_out3.PacketNr) * sizeof(message_out3.value)] , message_out3.value,sizeof(message_out3.value));
									//else
									//memcpy_letohl(&arr[arraypointer + message_out3.PacketNr * sizeof(message_out3.value)] , message_out3.value,size);
									paket_erreicht[letohl(message_out3.PacketNr)] = 1;
									size = size - sizeof(message_out3.value);
								}
							}
							else
							{
								retry = retry + 1;
								incRetryCounter();
								if (retry > k_max_retry_get)
								{
									D->fehlercode = k_error_tcp_error;
									return(1);
								}
								goto nochmal_empfangen2;
							}
						}
					}
					else
					{
						// Fehlerhandling falscher empfang
						retry = retry + 1;
						incRetryCounter();
						if (retry > k_max_retry_get)
						{
							D->fehlercode = k_error_tcp_error;
							return(1);
	}
						goto nochmal_empfangen2;
					}
				}
				else
				{
					// Fehlerhandling timeout
					retry = retry + 1;
					incRetryCounter();
					if (retry > k_max_retry_get)
					{
						D->fehlercode = k_error_tcp_timout_read;
						return(1);
					}
					if (anzahl_empfangen == 0) goto nochmal_senden;
					else value1 = max_paket_count ;
				}
			}
		}

		// Falls alles richtig
		fehlerblock = -1;
		for (value1 = 0; value1 < max_paket_count ; value1++)
		{
			{
				int helpi;
				helpi = last_size - value1 * sizeof(message_out3.value);
				if (helpi <= 0) helpi = last_size + 1;

				if (last_size - helpi >= 0)
				{
					if (paket_erreicht[value1] != 1)
					{
						fehlerblock = value1;
						value1 = max_paket_count;
					}
				}
			}
		}
		if (fehlerblock == -1)
		{
			message_out.status = htolel(k_ok);
			message_out.value[0] = htolel(max_paket_count);
			arraypointer = arraypointer + (last_size - size);
		}
		else
		{
			message_out.status = htolel(k_PaketFehler);
			message_out.value[0] = htolel(fehlerblock);
			size = last_size - fehlerblock * sizeof(message_out3.value);
			arraypointer = arraypointer + fehlerblock * sizeof(message_out3.value);
			if (fehlerblock == 0) {
				retry = retry + 1;
				incRetryCounter();
			}
			if (retry > k_max_retry_get)
			{
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}
		retry = 0;
		pNr++;
		message_out.paket_index= htolel(pNr);
		counti = counti + 1;
nochmal_senden:;

		if (first_empfang == 0)
		{
			message_out.status = htolel(k_PaketFehler);
			message_out.value[0] = 0;
			message_out.paket_index = 0;
			retry = retry + 1;
			incRetryCounter();
			if (retry > k_max_retry_get)
			{
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}

/*#ifdef _MYUDPTEST
		if (	dummy_zaehler == 7)
			dummy_zaehler  = 0;
		else
		{
#endif //_MYUDPTEST*/
			res = sendto(msocket, (char *) &message_out, sizeof(UDPR1), 0,m_serv_addr,sizeof(struct sockaddr));
			dummy_zaehler = dummy_zaehler + 1;
/*#ifdef _MYUDPTEST
		}
#endif //_MYUDPTEST*/
		anzahl_empfangen = 0;
	}

	D->fehlercode = 0;

	return(0);
}

// \param swap_bytes: if 0, just sent data as is; if 1, perform host-to-little-enian conversion
// TRUE Fehler FALSE ok
short setarray (unsigned char *arr, int32_t msize, int32_t typsize,SOCKET_TYPE msocket, struct sockaddr *server, int max_paket_count, struct timeval *tv, int swap_bytes)
{

	int32_t size,last_size,retry,zaehler,zaehler1,zaehler3;
	int32_t arraypointer;
	int32_t last_arraypointer;
	int32_t packet_nr;
	int fehlerblock;
	UDPR3 message_out3; // sent
	UDPR1 message_out; // received

	int value1,res,retry2,retry3,retry4;

	fd_set  my_fd_setl;

	zaehler = 0;
	zaehler1 = 1;
	zaehler3 = 0;

	fehlerblock = 0;
	arraypointer = 0;
	packet_nr = 1;
	retry = 0;
	retry2 = 0;
	size = msize * typsize; // Gesamtgroesse Datei

	while(size > 0)
	{

		last_arraypointer = arraypointer;
		last_size = size;
nochmal_senden:;
		for (value1 = 0; value1 < max_paket_count; value1++)
		{
			if (size > 0)
			{

				if (size > sizeof(message_out3.value)) {
					//printf("copying %d bytes from %p to %p", sizeof(message_out3.value), &arr[arraypointer + value1 * sizeof(message_out3.value)], message_out3.value);
					
					if (swap_bytes) {
						memcpy_htolel(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],sizeof(message_out3.value));
					} else {
					memcpy(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],sizeof(message_out3.value));

					}
				} else {
					//printf("copying %d bytes from %p to %p", size, &arr[arraypointer + value1 * sizeof(message_out3.value)], message_out3.value);
					if (swap_bytes) {
						memcpy_htolel(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],size); 
					} else {
					memcpy(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],size);
					} 
				}
				message_out3.PacketNr = htolel(value1);	// von 0 bis 4
				message_out3.paket_index = htolel(zaehler1);

//#ifdef _MYUDPTEST
//				zaehler3 = zaehler3 + 1;
//				if (	zaehler3 < 17)
//				{
//#endif //_MYUDPTEST
					res = sendto(msocket, (char *) &message_out3, sizeof(message_out3), 0, server, sizeof(struct sockaddr));
//#ifdef _MYUDPTEST
//				}
//				else
//					zaehler3 = 0;
//#endif //_MYUDPTEST
//				if (res == sizeof(message_out3))
//				{
					size = size - sizeof(message_out3.value);
			}
		}

		retry3 = 0;
		retry4 = 0;
		retry = 0;
		retry2 = 0;

nochmal_empfangen:;

		FD_ZERO(&my_fd_setl);
		FD_SET(msocket,&my_fd_setl);
		RESET_TIMEOUT (*tv)

		if (1 ==select(D->sockfd+1,&my_fd_setl,0,0 ,tv))
		{
			struct sockaddr dummy;
			socklen_t len_dummy;

			len_dummy = sizeof(dummy);
			res = recvfrom(msocket, (char *)&message_out, sizeof(message_out), 0, &dummy, &len_dummy);
			if (res == sizeof(message_out))
			{
				if ((size <= 0) && (letohl(message_out.status) == k_ok))
				{
					return(0); //False
				}

				if (letohl(message_out.paket_index) != zaehler1)
				{
					if (letohl(message_out.paket_index) == zaehler1-1)
					{
						retry4 = retry4 + 1;

						if (retry4 < 3)
							goto nochmal_empfangen;
						else
						{
							// Nochmal senden
							size = last_size;
							goto nochmal_senden;
						}
					}
					else
					{

						// Falls 1 Block und Falls paket_index = udp_paket_count

						retry4 = 0;

						if ((zaehler1 == 1) && ((unsigned long)(letohl(message_out.paket_index)) == D->udp_paket_count))
						{
							message_out.status = htolel(k_PaketFehler);
							message_out.value[0] = 0; // htolel(0);
							zaehler1 = 0;						// Neu 27.04.04  Damit es mehr als einmal Wiederholt wird !						}
						}
						else
							goto nochmal_empfangen;
					}
				}
				zaehler1 = zaehler1 + 1;
				if (letohl(message_out.status) != k_ok)
				{
					if (letohl(message_out.status) == k_PaketFehler)
					{
						arraypointer = last_arraypointer + letohl(message_out.value[0]) * sizeof(message_out3.value);
						size = last_size - letohl(message_out.value[0]) * sizeof(message_out3.value);
					}
					else
					{
						D->fehlercode = letohl(message_out.status);
						return(1);
					}
					retry = retry + 1;
					incRetryCounter();
					if (retry > k_max_retry)
					{
						D->fehlercode = k_error_tcp_error;
							return(1);
					}
				}
				else
					arraypointer = last_arraypointer + max_paket_count * sizeof(message_out3.value);

			}
			else
			{

				if ((size <= 0) && (retry2 == 2))
				{
					D->fehlercode = 0;
					return(0);
				}
				retry3 = retry3 + 1;
				incRetryCounter();
				if (retry3 < k_max_retry)  goto nochmal_empfangen;
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}
		else
		{
			// irgentwas schiefgegangen timeout erreicht
			retry3 = retry3 + 1;
			incRetryCounter();
			if (retry3 < k_max_retry)  goto nochmal_empfangen;
			D->fehlercode = k_error_tcp_timout_read;
			return(1);
		}

	}
	D->fehlercode = 0;
	return(0);
}
/*
	Data 	Pointer auf Array

	typ	 	Datentyp
		1 	16 Bit GANZZAHL
		2	32 BIT GANZZAHL
		5	32 BIT FLOAT
		6	double
	   20	VARIANT

	DataNo 	Datanummer
	StartIndex	erster Wert
	Count		Anzahl
	nadr	Linkaddresse
*/

void    NAME (Get_Data) (int32_t data_no, void *dest, adwin_data_type_t type, int32_t start_index, int32_t count) {
	// @@MB protocol depends on firmware version (must be >= 2.66)
	// and the amount of data to be transferred (must be > blocksize * 350)
	// old protocol: getDataClassic
	// new protocol: getDataSH (getData shifted handshake)

	if (count < 1) return;
	
	if ((D->ip_version < 266) || (count <= (5 * 350)))	//TODO: blocksize
		getDataClassic(data_no, dest, type,  start_index, count);
	else
		getDataSH(D->timeout, dest, type, (short)data_no, start_index, count);
}

#define g_QNAN_MASKE 0x7F800000

void test_float_buffer_qnan(void *buffer, int32_t count)
{
	int32_t *values;
	float *fvalue;
	int32_t lauf;
	
	values = (int32_t *)buffer;
	for (lauf = 0; lauf < count; lauf ++) {
		if ((values[lauf] & g_QNAN_MASKE) == g_QNAN_MASKE) {
			fvalue = (float *)buffer;
			fvalue[lauf] = FLT_MAX;
		}
	}
}

void getDataSH(int32_t timeout, void *data, short typ, short dataNo, int32_t startIndex, int32_t count)
{

	uint32_t status;
	short dsp_typ;
	s_getDataQueue getDataQueue;

	UDPH1 h1;			
	UDPR1 r1;			
	struct timeval my_time; 
	char *daten;
	char *datenSicherung;
	uint32_t helpi;
	int32_t blksize;
	int processorType;
	int startIndexT11, countT11;		// Vor- und Nachspann zum Verwerfen
	int alignedT11;						// Anzahl oder Startindex wegen Tiger verndert -> malloc und memcpy
	int ret;
	int myRetryCounter = 0;

	// TODO: abhngig von PC / Betriebssystem maxReq und minReq anpassen:
	unsigned short maxReq = 30;			// maximal offene Pakete
	unsigned short minReq = 6;			// minimal offene Pakete, wenn ereicht: neue Anforderung schicken

	if (count < 1) return;

	memset(&getDataQueue, 0, sizeof(getDataQueue));
	getDataQueue.requestNumber = 1;

	startIndexT11 = startIndex;
	countT11 = count;

	init_packet(&my_time, &h1);
	h1.timeout = htolel(D->timeout * 5);

 	h1.befehl   = htolel(k_GetDataSH);
 	h1.value[0] = htonl((dataNo<<16) | (maxReq & 0xFFFF));
	h1.value[1] = htonl(startIndex);
	h1.value[2] = htonl(count);
	h1.value[3] = 0;

	//if (do_udp_command(mutex,&h1,&r1,&my_time)) return 255;		// Fehlercode ist gesetzt
	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);//255 bei M.G.
	getDataQueue.packetIndex = D->udp_paket_count;
	getDataQueue.queueInsertPointer = 0;
	getDataQueue.queueDeletePointer = 0;
	getDataQueue.sockfd = D->sockfd;
	getDataQueue.serverAddr = (struct sockaddr *) &D->server;

#ifndef BUILD_LIBRARY
	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;
#else
	dsp_typ = typ;
#endif	
	//dsp_typ = (short)r1.value[0];
	processorType = ntohl(r1.value[1]);

	status = r1.status;

	if (dsp_typ == 0)	// Fehlerauswertung:
	{
		if (status == k_DataZuKlein) 
			D->fehlercode = k_Data_is_to_small;
		else
			D->fehlercode = status; // kann nur noch 1 oder 2 sein (To/FromLinkTimeout)

		return;
	}


	// Tiger? -> Startindex und Anzahl anpassen
	if (processorType == T11)
	{
		int dummy;
		
		dummy = ((startIndex - 1) & 3);
		startIndexT11 = startIndex - dummy;
		countT11 = 4 * ((count + dummy + 3) >> 2);
	}

	alignedT11 = ((startIndexT11 != startIndex) || (count != countT11)) ? 1 : 0;

	if ((dsp_typ != typ) || alignedT11)
	{
		//blksize = tcpip[(long)mutex->device].blocksize;
		blksize = 5;
		helpi = (countT11 / (350 * blksize)) + 1;
		helpi = helpi * (350 * blksize );
		daten = malloc(sizeof(int32_t) * helpi);
	}
	else
		daten = data;


	ret = getarray_shifted_handshake(&getDataQueue, timeout, countT11, (int32_t *)daten, minReq, maxReq, &myRetryCounter);
	
	for (; myRetryCounter; incRetryCounter(), myRetryCounter--);
	
	if (ret == -1)
	{
		if (dsp_typ != typ)
			free(daten);
		D->fehlercode = getDataQueue.errorCode;
		return;
	}

	datenSicherung = daten;

	// wenn startIndex verndert -> Zeiger verschieben
	if (startIndexT11 != startIndex)
	{
		daten += 4*(startIndex - startIndexT11);
	}


	if (dsp_typ == 5)
		test_float_buffer_qnan(daten, count);


#ifndef BUILD_LIBRARY
	// if the memory was copied by conversion, just free
	// dateSicherung. If no conversion was done, check alignedT11
	// to determine what to do with datenSicherung.
	if (Get_Data_do_convert(dsp_typ, daten, typ, data, count)) {
		free(datenSicherung);
	} else 
#else
	if (alignedT11) {
		memcpy(data, daten, count * sizeof(int32_t));
		free(datenSicherung);
	}
#endif	
}

void getDataClassic (int32_t data_no, void *Data, short type, int32_t start_index, int32_t count) {	
	short dsp_typ;
	char *daten;

	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	if (count < 1) return;//(1);//0 bei M.G.

	D->fehlercode = 0;

	// falls klein und Version > als
	// 28.01.02 nderungen wegen Jrgen Version von 300 auf 250 zurckgeasetzt,
	if (( count <= 5 * 350) && (D->ip_version >= 250)) {
		D->fehlercode = Get_Data_fastshort(data_no, Data, type, start_index, count);
		return;
	}

#ifndef BUILD_LIBRARY
	dsp_typ = (short)NAME (GetDataType) ((short)data_no);

	if (dsp_typ == 0) {
		if (D->fehlercode == 0) D->fehlercode = k_Data_is_to_small;
		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
		return;
	}
	
	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;
#else
	dsp_typ = type;
#endif	
	//daten = Get_Data_prepare_convert(dsp_typ, Data, typ, count);
	// todo: this is ugly, and should be replaced by Get_Data_prepare_convert
 	{
		uint32_t helpi;
		helpi = (count / (350 * 5)) + 1;
		helpi = helpi * (350 * 5 );
		daten = malloc(sizeof(int32_t) * helpi);
 	}

//udp_retry_get_data:
	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_GetData);
 	h1.value[0] = htolel(dsp_typ); // maybe todo: LONG instead of SHORT
 	h1.value[1] = htolel(data_no | (5 << 16));
 	h1.value[2] = htolel(start_index);
 	h1.value[3] = htolel(count);
 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;
 	if(getarray (daten, count, sizeof(int32_t), D->sockfd, (struct sockaddr *) &(D->server), 5, &my_time)) {
		free(daten);
		return;
	}

 	D->fehlercode = 0;
#ifndef BUILD_LIBRARY
	if (Get_Data_do_convert(dsp_typ, daten, type, Data, count) == 0)
		memcpy(Data, daten, count * sizeof(int32_t));
#else
	memcpy(Data, daten, count * sizeof(int32_t));
#endif	
	free(daten);
}

/* void getDataClassic (int32_t DataNo, void *Data, short typ, int32_t StartIndex, int32_t Count) */
/* {	 */
/* 	int32_t l1; */

/* 	int32_t  *pint; */
/* 	float *pfloat; */

/* 	int pointer,start,msize; */
/* 	short dsp_typ; */
/* 	char *daten; */

/* 	UDPH1 h1; */
/* 	UDPR1 r1; */
/* 	struct timeval my_time; */

/* 	if (Count < 1) return;//(1);//0 bei M.G. */

/* 	D->fehlercode = 0; */

/* 	// falls klein und Version > als */
/* 	// 28.01.02 nderungen wegen Jrgen Version von 300 auf 250 zurckgeasetzt, */
/* 	if (( Count <= 5 * 350) && (D->ip_version >= 250)) */
/* 	{ */
/* 		D->fehlercode = Get_Data_fastshort(DataNo, Data, typ, StartIndex, Count); */
/* 		return; */
/* 	} */

/* 	dsp_typ = (short)NAME (GetDataType) ((short)DataNo); */

/* 	if (dsp_typ == 0) */
/* 	{ */
/* 		if (D->fehlercode == 0) D->fehlercode = k_Data_is_to_small; */
/* 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small; */
/* 		return;//(1); //255 bei M.G. */
/* 	} */

/*  	if ((dsp_typ == 3) || (dsp_typ == 4)) */
/*  		dsp_typ = 2; */

/*  	//if (dsp_typ != typ) */
/*  	{ */
/* 		uint32_t helpi; */
/* 		int32_t blksize; */

/* 		blksize = 5; */
/* 		helpi = (Count / (350 * 5)) + 1; */
/* 		helpi = helpi * (350 * 5 ); */
/* 		daten = malloc(sizeof(long) * helpi); */
/*  		//daten = malloc(sizeof(long) * Count); */
/*  	} */
/*  	//else														//Neu */
/*  		//daten = Data; */

/* //udp_retry_get_data: */

/*  	msize = sizeof(long); */
/*  	pointer = 0; */
/*  	start = StartIndex; */

/* 	init_packet(&my_time, &h1); */

/*  	h1.befehl   = htolel(k_GetData); */
/*  	h1.value[0] = htolel(dsp_typ); */
/*  	h1.value[1] = htolel(DataNo | (5 << 16)); */
/*  	h1.value[2] = htolel(start); */
/*  	h1.value[3] = htolel(Count); */
/*  	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);//255 bei M.G. */
/*  	if(getarray (daten, Count, msize, D->sockfd, (struct sockaddr *) &(D->server), 5, &my_time)) */
/* 	{ */
/* 		free(daten); */
/* 		return;// 255; */
/* 	} */

/*  	D->fehlercode = 0; */

/*  	if (dsp_typ != typ) */
/*  	{ */
/*  		switch (typ) */
/*  		{ */
/* 			//TODO: case 1, 6, 20 */
/*  			case 2:				// 32bit int */
/*  				pint = Data; */
/*  				if (dsp_typ == 5) */
/*  				{ */
/*  					pfloat =  (float *)  daten; */
/*  					for (l1=0; l1<Count  ; l1++) */
/*  					{ */
/*  						pint[l1] = (long)(pfloat [l1]); */
/* 					} */
/*  				} */
/*  				break; */
/*  			case 5:				// 32bit float */
/*  				pfloat = Data; */
/*  				if (dsp_typ == 2) */
/*  				{ */
/*  					pint =  (int32_t *)  daten; */
/*  					for (l1=0; l1<Count  ; l1++) */
/*  					{ */
/*  						pfloat[l1] = (float)(pint[l1]); */
/* 					} */
/*  				} */
/*  				break; */
/* 			default:; */
/*  		} */
/*  		free(daten); */
/*  	} */
/* 	else */
/* 	{ */
/* 		memcpy(Data,daten,Count * sizeof(long)); */
/* 		free(daten); */
/* 	} */
/* } */

short Get_Data_fastshort(int32_t DataNo,void *Data, short typ,  int32_t StartIndex, int32_t count)
{
	short dsp_typ = 0;
	char *daten;
	UDPH1 h1;
	UDPR4 r4;
	struct timeval my_time;

	int res,versuche,blockcount;

	init_packet(&my_time, &h1);

	versuche = 0;

	blockcount = ((count - 1) / 350) + 1 ;

	daten = malloc(sizeof(int32_t) * blockcount * 350);

	if (daten == NULL)
	{
		return (k_Data_is_to_small);
	}

	h1.befehl   = htolel(k_GetDatasmall);
	h1.value[0] = htolel(dsp_typ);
	h1.value[1] = htolel(DataNo | (blockcount << 16));
	h1.value[2] = htolel(StartIndex);
	h1.value[3] = htolel(count);

//Get_Data_fastshort_nochmal:;
	res = do_udp_command_nx4(D->sockfd, &h1, &r4, &(D->server), &my_time, blockcount, daten, &dsp_typ);

	if (res < blockcount)
	{
		// hier stimmt was nicht
		// Dumme Fehlerlsung besser wre es nun nur noch den Rest abzuholen.

/*		if ((fehlercode == k_Data_is_to_small) || (fehlercode == k_error_tcp_password)|| (fehlercode == k_timeout_to_link)|| (fehlercode == k_timeout_from_link))
		{
			return(255);
		}*/
/*		if (versuche < 5)							// da schon in do_udp_kommand_nx4
		{
			versuche = versuche + 1;
			goto Get_Data_fastshort_nochmal;		// so net!!!
		}*/
		free(daten);
		return((short)(D->fehlercode));		// wird vom do_udp_command_nx4 bereits richtig gesetzt
	}

	if (dsp_typ == 0)
	{
		free(daten);
		return (k_Data_is_to_small);
	}

	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;

#ifndef BUILD_LIBRARY
	if (Get_Data_do_convert(dsp_typ, daten, type, Data, count) == 0)
		memcpy(Data, daten, count * sizeof(int32_t));
#else
	memcpy(Data, daten, count * sizeof(int32_t));
#endif
	free(daten);

	return(0);
}


void    NAME (Set_Data) (int32_t data_no, void *src, adwin_data_type_t type, 
			 int32_t start_index, int32_t count) {
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	short res;
	int pointer,start,mysize,retry;
	short dsp_typ;
	unsigned char *daten;
	int32_t paket_index_dieses_Befehls;

	if (count < 1) return;

	D->fehlercode = 0;

#ifndef BUILD_LIBRARY
	dsp_typ = (short)NAME (GetDataType) ((short)data_no);
	
	if (dsp_typ == 0) {
 		if (D->fehlercode == 0) 
			D->fehlercode = k_Data_is_to_small;
 		if (D->fehlercode == k_DataZuKlein) 
			D->fehlercode = k_Data_is_to_small;
 		return;
 	}

	if ((dsp_typ == ADWIN_SYSTEM_SHORT) || (dsp_typ == ADWIN_SYSTEM_INTEGER))
 		dsp_typ = ADWIN_SYSTEM_LONG;

	daten = Set_Data_convert(type, src, dsp_typ, count);
#else
	dsp_typ = type;
	daten = src;
#endif

	retry = 0;
	pointer = 0;
	start = start_index;

	mysize = sizeof(int32_t);

	init_packet(&my_time, &h1);

 	h1.befehl   = htolel(k_SetData);
 	h1.value[0] = htolel(dsp_typ); // this is now mapped to LONG, but was mapped to short before... TODO
 	h1.value[1] = htolel(data_no | (5 << 16));
 	h1.value[2] = htolel(start);
 	h1.value[3] = htolel(count);

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
 	{
 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
		if (src != daten) free(daten);
 		return;
 	}

 	if (letohl(r1.status) != k_ok)
 	{
 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
		if (src != daten) free(daten);
 		return;
 	}

	paket_index_dieses_Befehls = D->udp_paket_count;			// Neu 06.04.04
 	if (setarray(daten, count, sizeof(int32_t),D->sockfd, (struct sockaddr *) &(D->server) , 5, &my_time, 1))
 	{
		if (src != daten) free(daten);
 		return;
 	}
 	if (src != daten) free(daten);

udp_retry:;

 	res = recvfromServer(D->sockfd, &r1, &my_time);

	// Hier mu berprft werden ob Server timeout sendet!!
	if (k_ok != res) {
		// Fehlerbehandlung
		
		if (res > 0) {
			D->fehlercode = res;
			return;
		}
		
		D->fehlercode = k_error_tcp_error;
		
		if (res == k_UDPtimeout) {
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry == 0) {
			retry = 1;
			goto udp_retry;
		}
		// Falls alles schiefgegangen ist !!		// Neu 06.04.04
		if (D->ip_version >= 264 ||
		    ((D->ip_version >= 253) && (D->ip_version < 260))) {
			init_packet(&my_time, &h1);
			
			h1.befehl   = htolel(k_SetDataLastStatus);
			h1.value[0] = htolel(paket_index_dieses_Befehls);
			h1.value[1] = 0;
			h1.value[2] = 0;
			
			if (do_udp_command (D->sockfd, &h1, &r1, &(D->server), &my_time)) {
				D->fehlercode = k_error_tcp_error;
				return;
			}
			if (letohl(r1.value[0]) == 1) {
				// war ok
				D->fehlercode = 0;
				return;
			} else {
				D->fehlercode = k_error_tcp_error;
			}
		}	// Neu 06.04.04 Ende				
		return;
	}
	D->fehlercode = 0;
	
	if (k_ok == res) {
		if (letohl(r1.status) != k_ok) {
			//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
			switch (letohl(r1.status)) {
			case 1:
				D->fehlercode = k_timeout_to_link;
				break;
			case 2:
				D->fehlercode = k_timeout_from_link;
				break;
			case 5:
				D->fehlercode = k_timeout_to_link;
				break;
			case -1:
			case -10:
			case -15:
			case -20:
			case -25:
			case -26:
			case -30:
			case -33:
			case -40:
			case -41:
				D->fehlercode = k_error_tcp_error;
				break;
			case -5:
				D->fehlercode = k_error_tcp_password;
				break;
				
			case -31:
				D->fehlercode = k_Data_is_to_small;
				break;

			case -32:
				D->fehlercode = k_Fifo_to_less_data;
				break;
			}
			return;
		}
	}
}

/* void    NAME (Set_Data) (int32_t data_no, void *src, adwin_data_type_t type,  */
/* 			 int32_t start_index, int32_t count) { */
/* 	UDPH1 h1; */
/* 	UDPR1 r1; */
/* 	struct timeval my_time; */

/* 	short res; */
/* 	int pointer,start,mysize,retry; */
/* 	short dsp_typ; */
/* 	unsigned char *daten; */
/* 	int32_t paket_index_dieses_Befehls; */

/* 	if (count < 1) return; */

/* 	D->fehlercode = 0; */

/* 	dsp_typ = (short)NAME (GetDataType) ((short)data_no); */
	
/* 	if (dsp_typ == 0) */
/*  	{ */
/*  		if (D->fehlercode == 0)  */
/* 			D->fehlercode = k_Data_is_to_small; */
/*  		if (D->fehlercode == k_DataZuKlein)  */
/* 			D->fehlercode = k_Data_is_to_small; */
/*  		return; */
/*  	} */

/* 	if ((dsp_typ == ADWIN_SYSTEM_INTEGER) || (dsp_typ == ADWIN_SYSTEM_LONG)) */
/*  		dsp_typ = ADWIN_SYSTEM_SHORT; */

/*  	if (dsp_typ != type) { */
/*  		daten = malloc(sizeof(long) * count); */
/*  		switch (type) { */
/* 		case ADWIN_SHORT: */
/* 			if (dsp_typ == ADWIN_SYSTEM_SHORT) { */
/* 				CAST(short, src, int32_t, daten, count); */
/* 			} */
/* 			if (dsp_typ == ADWIN_SYSTEM_FLOAT) { */
/* 				CAST(short, src, float, daten, count); */
/* 			} */
/*  			break; */
/* 		case ADWIN_LONG: */
/*   			if (dsp_typ == ADWIN_SYSTEM_FLOAT) { */
/* 				CAST(int32_t, src, float, daten, count); */
/*   			} */
/*   			break; */
/* 		case ADWIN_FLOAT: */
/* 			if (dsp_typ == ADWIN_SYSTEM_SHORT) { */
/* 				CAST(float, src, short, daten, count); */
/* 			} */
/* 			break; */
/*   		} */
/* 	} */
/* 	else */
/* 		daten = (unsigned char *)src; */

/* 	retry = 0; */
/* 	pointer = 0; */
/* 	start = start_index; */

/* 	mysize = sizeof(long); */

/* 	init_packet(&my_time, &h1); */

/*  	h1.befehl   = htolel(k_SetData); */
/*  	h1.value[0] = htolel(dsp_typ); */
/*  	h1.value[1] = htolel(data_no | (5 << 16)); */
/*  	h1.value[2] = htolel(start); */
/*  	h1.value[3] = htolel(count); */

/*  	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) */
/*  	{ */
/*  		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small; */
/*  		if (dsp_typ != type) free(daten); */
/*  		return; */
/*  	} */

/*  	if (letohl(r1.status) != k_ok) */
/*  	{ */
/*  		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small; */
/*  		if (dsp_typ != type) free(daten); */
/*  		return; */
/*  	} */

/* 	paket_index_dieses_Befehls = D->udp_paket_count;			// Neu 06.04.04 */
/*  	if (setarray(daten, count, sizeof(long),D->sockfd, (struct sockaddr *) &(D->server) , 5, &my_time, 1)) */
/*  	{ */
/*  		if (dsp_typ != type) free(daten); */
/*  		return; */
/*  	} */
/*  	if (dsp_typ != type) free(daten); */

/* udp_retry:; */

/*  	res = recvfromServer(D->sockfd, &r1, &my_time); */

/* 	// Hier mu berprft werden ob Server timeout sendet!! */
/* 			if (k_ok != res) */
/* 			{ */
/* 				// Fehlerbehandlung */

/* 				if (res > 0) */
/* 				{ */
/* 					D->fehlercode = res; */
/* 					return;//(255); */
/* 				} */

/* 				D->fehlercode = k_error_tcp_error; */

/* 				if (res == k_UDPtimeout) */
/* 				{ */
/* 					D->fehlercode = k_error_tcp_timout_read; */
/* 				} */
/* 				if (retry == 0) */
/* 				{ */
/* 					retry = 1; */
/* 					goto udp_retry; */
/* 				} */
/* 				// Falls alles schiefgegangen ist !!		// Neu 06.04.04 */
/* 				if (D->ip_version >= 264 || */
/* 					((D->ip_version >= 253) && (D->ip_version < 260))) */
/* 				{ */
/* 					init_packet(&my_time, &h1); */

/* 					h1.befehl   = htolel(k_SetDataLastStatus); */
/* 					h1.value[0] = htolel(paket_index_dieses_Befehls); */
/* 					h1.value[1] = 0; */
/* 					h1.value[2] = 0; */
				
/* 					if (do_udp_command (D->sockfd, &h1, &r1, &(D->server), &my_time)) */
/* 					{ */
/* 						D->fehlercode = k_error_tcp_error; */
/* 						return; */
/* 					} */
/* 					if (letohl(r1.value[0]) == 1) */
/* 					{ */
/* 						// war ok */
/* 						D->fehlercode = 0; */
/* 						return; */
/* 					}					 */
/* 					else */
/* 					{ */
/* 						D->fehlercode = k_error_tcp_error; */
/* 					} */
/* 				}									// Neu 06.04.04 Ende				 */
/* 				return;//(255); */
/* 			} */
/* 			D->fehlercode = 0; */

/* 			if (k_ok == res) */
/* 			{ */
/* 				if (letohl(r1.status) != k_ok) */
/* 				{ */
/* 					//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows */
/* 						switch (letohl(r1.status)) */
/* 						{ */
/* 							case 1: */
/* 								D->fehlercode = k_timeout_to_link; */
/* 							break; */
/* 							case 2: */
/* 								D->fehlercode = k_timeout_from_link; */
/* 							break; */
/* 							case 5: */
/* 								D->fehlercode = k_timeout_to_link; */
/* 							break; */
/* 							case -1: */
/* 							case -10: */
/* 							case -15: */
/* 							case -20: */
/* 							case -25: */
/* 							case -26: */
/* 							case -30: */
/* 							case -33: */
/* 							case -40: */
/* 							case -41: */
/* 								D->fehlercode = k_error_tcp_error; */
/* 							break; */
/* 							case -5: */
/* 								D->fehlercode = k_error_tcp_password; */
/* 							break; */

/* 							case -31: */
/* 								D->fehlercode = k_Data_is_to_small; */
/* 							break; */

/* 							case -32: */
/* 								D->fehlercode = k_Fifo_to_less_data; */
/* 							break; */
/* 						} */
/* 					return ;//255; */
/* 				} */
/* 			} */
/* 	return ;//(res) ; */
/* } */


long get_filesize(FILE *fd)
{
	long size;

	fseek(fd, 0, SEEK_END);
	size = ftell(fd);
	rewind(fd);

	return (size);
}


int32_t NAME (Free_Mem_Byte) (short typ)
{
        int32_t ret = NAME (Free_Mem) (typ);

        if (typ == 1) {
		processor_type_t processor = NAME (Processor_Type)();

                if ((processor == PROCESSOR_T9) || 
		    (processor == PROCESSOR_T10))
                        return (6 * ret);
        }

        return (4*ret);
}

void NAME (Get_Memory_dsp) (int32_t base_addr, int32_t count, int32_t *dest)
{
	UDPH1 h1;
	UDPR2 r2;
	struct timeval my_time;
	int32_t done;                      // longs already transferred
	int32_t count_packet; 		// longs to get with next packet

	init_packet(&my_time, &h1);

	done = 0;
	while(count > 0)
	{
		if (count * sizeof(int32_t) > sizeof(r2.value)) // need more than one packet
			count_packet = sizeof(r2.value) / sizeof(int32_t);
		else // fits into one packet
			count_packet = count;

		h1.befehl   = htolel(k_GetMemoryDSP);
		h1.value[0] = htolel(base_addr + done);
		h1.value[1] = htolel(count_packet);

		if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;

		memcpy_letohl(&dest[done], r2.value, count_packet * sizeof(int32_t));

		count -= count_packet;
		done += count_packet;
	}				
}

/************Version, Prozessor**************************************************************/

int32_t intern_Get_Version_TcpIp() {

	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	PRINTDEBUG(DBG_GENERIC_DRV, ("intern_Get_Version_TcpIp()\n"));

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_armVersion);
	h1.value[0] = htolel(k_armVersion);
	h1.value[1] = 0;
	h1.value[2] = 0;
	h1.value[3] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) {
		PRINTDEBUG(DBG_GENERIC_DRV, ("intern_Get_Version_TcpIp() returned 0 (first)\n"));
		return 0;
	}

	D->fehlercode = 0;
	PRINTDEBUG(DBG_GENERIC_DRV, ("intern_Get_Version_TcpIp() returned %d\n", 
		   letohl(r1.value[0])));

	return(letohl(r1.value[0]));
}

int NAME (Clear_Data) (int32_t DataNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	init_packet(&my_time, &h1);

	h1.befehl   = htolel(k_3plus1);
	h1.value[0] = htolel(k_Clear_Data);
	h1.value[1] = htolel(DataNo);
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time ))
	{
		return 0;
	}

	return 1;
}
