#ifndef __LIBADWIN_H__
#define __LIBADWIN_H__

#include <sys/cdefs.h>
#include <sys/types.h>

#include <inttypes.h>

#include "libadwin_compat.h"
#include "libadwin_undocumented.h"
#include "libadwin_scilab.h"

#ifdef __cplusplus
extern "C" {
#endif

extern void    Set_DeviceNo(int32_t DeviceNo);

extern void    Boot(const char *filename);
extern int32_t Test_Version(void);
extern processor_type_t Processor_Type(void);
extern int32_t Workload(int priority);
extern int32_t Free_Mem (mem_spec_t mem_spec);

extern void    Load_Process(const char *filename);
extern void    Start_Process(int32_t process_no);
extern void    Stop_Process(int32_t process_no);
extern void    Clear_Process(int32_t process_no);
extern int32_t Process_Status(int32_t process_no);
extern void    Set_Processdelay(int32_t process_no, int32_t value);
extern int32_t Get_Processdelay(int32_t process_no);

extern void    Set_Par(int32_t index, int32_t value);
extern int32_t Get_Par(int32_t index);
extern void    Get_Par_Block(int32_t dest[], int32_t start_index, int32_t count);
extern void    Get_Par_All(int32_t dest[]);

extern void    Set_FPar(int32_t index, float value);
extern float   Get_FPar(int32_t index);
extern void    Get_FPar_Block(float dest[], int32_t start_index, int32_t count);
extern void    Get_FPar_All(float dest[]);
extern void    Get_FPar_Block_Double(double dest[], 
				     int32_t start_index, int32_t count);
extern void    Get_FPar_All_Double(double dest[]);

extern int32_t Data_Length(int data_no);

extern void    SetData_Long(int data_no, int32_t src[], 
			    int32_t start_index, int32_t count);
extern void    GetData_Long(int data_no, int32_t dest[],
			    int32_t start_index, int32_t count);
extern void    SetData_Float(int data_no, float src[],
			     int32_t start_index, int32_t count);
extern void    SetData_Double(int data_no, double src[], 
			      int32_t start_index, int32_t count);
extern void    GetData_Float(int data_no, float dest[], 
			     int32_t start_index, int32_t count);
extern void    GetData_Double(int data_no, double dest[], 
			      int32_t start_index, int32_t count);
extern void    Data2File(const char *filename, int data_no, 
			 int32_t start_index, int32_t count, int32_t mode);

extern int32_t Fifo_Empty(int fifo_no);
extern int32_t Fifo_Full(int fifo_no);
extern void    Fifo_Clear(int fifo_no);
extern void    SetFifo_Long(int fifo_no, int32_t src[], int32_t count);
extern void    GetFifo_Long(int fifo_no, int32_t dest[], int32_t count);
extern void    SetFifo_Float(int fifo_no, float src[], int32_t count);
extern void    GetFifo_Float(int fifo_no, float dest[], int32_t count);
extern void    SetFifo_Double(int fifo_no, double src[], int32_t count);
extern void    GetFifo_Double(int fifo_no, double dest[], int32_t count);

extern void    SetData_String(int data_no, char *src);
extern int     GetData_String(int data_no, char *dest, int32_t max_count);
extern int32_t String_Length(int data_no);

extern int32_t Get_Last_Error(void);
extern char*   Get_Last_Error_Text(int32_t Errno); // errno is already defined

extern int16_t Get_Known_DeviceNo(int32_t devices[], int32_t *count_devices);

extern const char* Get_Library_Version();

extern char *Get_Password();
	

#ifdef __cplusplus
}
#endif

#endif /* __LIBADWIN_H__ */
