****************************************
*	 Prozenummer = 1
*	 Delay = 1000
*	 Eventsource = 0
*	 Number of Loops = 0
*	 Prioritt = 0
*	 Version = 1
*	 FastStop = 0
*	 AdbasicVersion = 4000001
*	 ATSRAM = 0
*	 OPT_LEVEL = 1
*	 SAVECOMPIL = 0
****************************************
'###########################################################################
'# The process Pro_DMO6 is a digital PID-controller.
'# The controller operates with float variables who's coefficients
'# have to be computed on the PC and then transferred to the ADwin-system.
'# PAR_1  = setpoint in digits
'# FPAR_2 = P (gain, proportional factor)
'# FPAR_3 = I (integration time)
'# FPAR_4 = D (derivative time)
'# PAR_5  = buffer index for the control deviation
'# PAR_6  = cycle-time in 25ns steps (T9. T10 only)
'# PAR_9  = flag for new setpoint definition by PC
'# Attention: Prior to start the program the controller parameters
'#            have to be set to the correct value!
'#            Make sure that FPAR_3 is not zero!!!
'#
'# Platform: ADwin-Pro
'#              with AIn8/16 Rev B (or greater) (set to CARD ADDRESS 1)
'#              and AOut8/16 (set to CARD ADDRESS 1)
'###########################################################################

#include adwpad.inc ' Include-file for Pro-system 
#include adwpda.inc

#DEFINE offset 32768					'0V output

DIM DATA_1[4000] AS LONG
DIM av, cd, cdo, sum AS LONG
DIM diff AS FLOAT

INIT:
  sum = 0						'initial value of integral part
  cd = ADC(1,1)			'initial value of control deviation (cd) & MUX to Ch-#1
  PAR_5 = 1					'reset array index
  IF (FPAR_3 < 1E3) THEN FPAR_3 = 1E3	'check min. of integration time
  IF (PAR_6 < 4E4) THEN PAR_6 = 4E4		'allow cycle-times >= 1ms
  GLOBALDELAY = PAR_6									'set cycle-time

EVENT:
  av = FPAR_2 * (cd + sum / FPAR_3 + diff * FPAR_4)	'compute actuating value
  START_CONV(1)												'start conversion ADC
  DAC(1, 1, av + offset)							'output actuating value at DAC-#1
  cdo = cd														'keep control deviation in mind
  WAIT_EOC(1)													'wait until end-of-conversion of ADC
  cd = PAR_1 - READADC(1)							'compute control deviation
  FPAR_9 = FPAR_9 * 0.99 + cd * 0.01	'mean value of control deviation
  sum = sum + cd											'calculate integral
  IF (sum > 2E6) THEN sum = 2E6				'positive limit of integral
  IF (sum < -2E6) THEN sum = -2E6			'negative limit of integral
  diff = (cd - cdo)										'calculate deviation difference
  DATA_1[PAR_5] = cd									'write control deviation in a buffer
  INC PAR_5														'increment buffer index
  IF (PAR_5 >= 4000) THEN							'4000 samples done?
    ACTIVATE_PC							            'only for use with TestPoint
  	PAR_10 = 1	   					            'set End-flag
    PAR_5 = 1														'reset buffer index
  ENDIF

FINISH:
  DAC(1, 1, offset)			              'analog output #1 to 0V